TOOL_ELEMENTS = {
    "search_the_whole_bible": "search:::",
    "compare_bible_translations": "verses:::KJV,LEB,NET,OHGB,OHGBi:::",
    "retrieve_bible_cross_references": "xrefs:::",
    "retrieve_hebrew_or_greek_bible_verses": "verses:::OHGB:::",
    "retrieve_interlinear_hebrew_or_greek_bible_verses": "verses:::OHGBi:::",
    "retrieve_bible_verses": "verses:::",
    "retrieve_verse_morphology": "morphology:::",
    "retrieve_bible_chapter": "chapter:::",
    "read_bible_commentary": "commentary:::",
    "refine_bible_translation": {'system': 'bible/translate'},
    "write_pastor_prayer": {'system': 'bible/pray'},
    "ask_theologian": {'system': 'bible/theologian'},
    "quote_bible_verses": {'system': 'bible/quote'},
    "anyalyze_psalms": {'system': 'bible/david'},
    "ask_pastor": {'system': 'bible/billy'},
    "ask_bible_scholar": {'system': 'bible/scholar'},
    "write_bible_perspectives": {'system': 'bible/perspective'},
    "explain_bible_meaning": {'instruction': 'bible/meaning', 'system': 'auto'},
    "write_new_testament_historical_context": {'instruction': 'bible/nt_context', 'system': 'auto'},
    "write_bible_questions": {'instruction': 'bible/questions', 'system': 'auto'},
    "write_bible_devotion": {'instruction': 'bible/devotion', 'system': 'auto'},
    "translate_hebrew_bible_verse": {'instruction': 'bible/translate_hebrew', 'system': 'auto'},
    "write_bible_location_study": {'instruction': 'bible/location', 'system': 'auto'},
    "translate_greek_bible_verse": {'instruction': 'bible/translate_greek', 'system': 'auto'},
    "identify_bible_keywords": {'instruction': 'bible/keywords', 'system': 'auto'},
    "study_old_testament_themes": {'instruction': 'bible/ot_themes', 'system': 'auto'},
    "study_new_testament_themes": {'instruction': 'bible/nt_themes', 'system': 'auto'},
    "write_old_testament_highlights": {'instruction': 'bible/ot_highligths', 'system': 'auto'},
    "write_bible_prayer": {'instruction': 'bible/prayer', 'system': 'auto'},
    "write_short_bible_prayer": {'instruction': 'bible/short_prayer', 'system': 'auto'},
    "write_bible_character_study": {'instruction': 'bible/character', 'system': 'auto'},
    "write_bible_thought_progression": {'instruction': 'bible/flow', 'system': 'auto'},
    "quote_bible_promises": {'instruction': 'bible/promises', 'system': 'auto'},
    "write_bible_chapter_summary": {'instruction': 'bible/chapter_summary', 'system': 'auto'},
    "interpret_old_testament_verse": {'instruction': 'bible/ot_meaning', 'system': 'auto'},
    "expound_bible_topic": {'instruction': 'bible/topic', 'system': 'auto'},
    "write_bible_theology": {'instruction': 'bible/theology', 'system': 'auto'},
    "study_bible_themes": {'instruction': 'bible/themes', 'system': 'auto'},
    "write_bible_canonical_context": {'instruction': 'bible/canon', 'system': 'auto'},
    "write_bible_related_summary": {'instruction': 'bible/summary', 'system': 'auto'},
    "interpret_new_testament_verse": {'instruction': 'bible/nt_meaning', 'system': 'auto'},
    "write_new_testament_highlights": {'instruction': 'bible/nt_highlights', 'system': 'auto'},
    "write_bible_applications": {'instruction': 'bible/application', 'system': 'auto'},
    "write_bible_book_introduction": {'instruction': 'bible/introduce_book', 'system': 'auto'},
    "write_old_testament_historical_context": {'instruction': 'bible/ot_context', 'system': 'auto'},
    "write_bible_outline": {'instruction': 'bible/outline', 'system': 'auto'},
    "write_bible_insights": {'instruction': 'bible/insights', 'system': 'auto'},
    "write_bible_sermon": {'instruction': 'bible/sermon', 'system': 'auto'},
}


from nicegui import ui, app
from biblemateweb import get_translation, VerseEventObj
from biblemateweb.fx.bible import *
from biblemateweb.fx.original import *
from biblemateweb.fx.cloud_index_manager import get_drive_service, CloudIndexManager
from biblemateweb.js.sync_scrolling import *
from biblemateweb.data.cr_books import cr_books
from biblemateweb.data.lexical_data import lexical_data
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
import re, os


def bible_translation(gui=None, b=1, c=1, v=1, area=1, tab1=None, tab2=None, title="", **_):

    book_note = False
    verses_with_notes = []
    token = app.storage.user.get('google_token', "")
    if token:
        service = get_drive_service(token)
        index_mgr = CloudIndexManager(service)
        verses_with_notes = index_mgr.get_chapter_notes_verselist(b, c)
        if f"{b}_0_0" in index_mgr.data:
            book_note = True

    dummy_label1 = None
    dummy_label2 = None

    db = getBiblePath(title)
    if not os.path.isfile(db):
        return None

    bible_selector = BibleSelector(on_version_changed=gui.change_area_1_bible_chapter if area == 1 else gui.change_area_2_bible_chapter, on_book_changed=gui.change_area_1_bible_chapter if area == 1 else gui.change_area_2_bible_chapter, on_chapter_changed=gui.change_area_1_bible_chapter if area == 1 else gui.change_area_2_bible_chapter, on_verse_changed=change_bible_chapter_verse)

    def strong_to_lex(match):
        lexical_entry = match.group(1)
        if lexical_entry in lexical_data:
            return rf'<ref data-word="{lexical_entry}" class="tooltip-word">{lexical_data[lexical_entry][0]}'
        elif lexical_entry+"a" in lexical_data:
            return rf'<ref data-word="{lexical_entry}" class="tooltip-word">{lexical_data[lexical_entry+"a"][0]}'
        return match.group(0)

    def open_tool(tool):
        nonlocal area, gui
        if area == 2:
            gui.select_empty_area2_tab()
        gui.load_area_2_content(title=tool, sync=False)

    def lex(event):
        nonlocal gui
        lexical_entry, *_ = event.args
        app.storage.user['tool_query'] = lexical_entry
        open_tool("Lexicons")

    def wd(event):
        nonlocal gui
        lexical_entry, *_ = event.args
        app.storage.user['tool_query'] = lexical_entry
        open_tool("Lexicons")

    def note(event):
        nonlocal gui, area
        app.storage.user['tool_book_number'], app.storage.user['tool_chapter_number'], app.storage.user['tool_verse_number'], *_ = event.args
        open_tool("Notes")

    def book_studies(event):
        nonlocal gui, db, dummy_label1, area, verses_with_notes, book_note, bible_selector
        b, c, v = event.args
        with bible_selector.verse_select:
            gui.open_book_context_menu(db, b, c, v, book_note)

    def chapter_studies(event):
        nonlocal gui, db, dummy_label1, area, verses_with_notes, bible_selector
        b, c, v = event.args
        with bible_selector.verse_select:
            gui.open_chapter_context_menu(db, b, c, v, ("0" in verses_with_notes))

    def luV1_m(event):
        nonlocal bible_selector, gui, db, verses_with_notes
        b, c, v = event.args
        with bible_selector.verse_select:
            gui.open_verse_context_menu(db, b, c, v, (str(v) in verses_with_notes))

    def luV2_m(event):
        nonlocal bible_selector, gui, db, verses_with_notes
        b, c, v = event.args
        with bible_selector.verse_select:
            gui.open_verse_context_menu(db, b, c, v, (str(v) in verses_with_notes))

    def luV1(event):
        nonlocal bible_selector, gui, db, dummy_label1, verses_with_notes
        b, c, v = event.args
        bible_selector.verse_select.value = v
        gui.update_active_area1_tab_records(v=v)
        with dummy_label1:
            gui.open_verse_context_menu(db, b, c, v, (str(v) in verses_with_notes))

    def luV2(event):
        nonlocal bible_selector, gui, db, dummy_label2, verses_with_notes
        b, c, v = event.args
        bible_selector.verse_select.value = v
        gui.update_active_area2_tab_records(v=v)
        with dummy_label2:
            gui.open_verse_context_menu(db, b, c, v, (str(v) in verses_with_notes))

    def cr(event):
        nonlocal gui
        b, c, v, *_ = event.args
        b = cr_books.get(b, b)
        if app.storage.user["sync"]:
            gui.change_area_1_bible_chapter(None, b, c, v) if area == 1 else gui.change_area_2_bible_chapter(None, b, c, v)
        else:
            gui.change_area_2_bible_chapter(None, b, c, v) if area == 1 else gui.change_area_1_bible_chapter(None, b, c, v)

    def bcv(event):
        nonlocal gui, area
        b, c, v, *_ = event.args
        if app.storage.user["sync"]:
            gui.change_area_1_bible_chapter(None, b, c, v) if area == 1 else gui.change_area_2_bible_chapter(None, b, c, v)
        else:
            gui.change_area_2_bible_chapter(None, b, c, v) if area == 1 else gui.change_area_1_bible_chapter(None, b, c, v)

    ui.on('lex', lex)
    ui.on('note', note)
    ui.on('luV1', luV1)
    ui.on('luV2', luV2)
    ui.on('wd', wd)
    ui.on('bcv', bcv)
    ui.on('cr', cr)

    content = getBibleChapter(db, b, c)

    # Fix known issues
    content = content.replace("<br<", "<br><")
    content = content.replace("<heb> </heb>", "<heb>&nbsp;</heb>")

    # add tooltip
    # OHGB, OHGBi
    if "</heb>" in content:
        content = re.sub('(<heb id=")(.*?)"', r'\1\2" data-word="\2" class="tooltip-word"', content)
    elif "</grk>" in content:
        content = re.sub('(<grk id=")(.*?)"', r'\1\2" data-word="\2" class="tooltip-word"', content)
    # study notes
    if "<ref onclick='bn(" in content:
        content = re.sub(rf'''<ref onclick='bn\(([0-9]+?),[ ]*?([0-9]+?),[ ]*?([0-9]+?),[ ]*?"(.*?)"\)'>''', rf'<ref data-word="bn,{title},\1,\2,\3,\4" class="tooltip-word">', content)
    elif '<ref onclick="bn(' in content:
        content = re.sub(rf'''<ref onclick="bn\(([0-9]+?),[ ]*?([0-9]+?),[ ]*?([0-9]+?),[ ]*?'(.*?)'\)">''', rf'<ref data-word="bn,{title},\1,\2,\3,\4" class="tooltip-word">', content)
    # Strong's numbers
    if "<ref onclick='lex(" in content:
        content = re.sub(r'''<ref onclick='lex\("(.*?)"\)'>\1''', strong_to_lex, content)
    elif '<ref onclick="lex(' in content:
        content = re.sub(r'''<ref onclick="lex\('(.*?)'\)">\1''', strong_to_lex, content)

    # convert verse link, like '<vid id="v19.117.1" onclick="luV(1)">'
    content = re.sub(r'<vid id="v([0-9]+?)\.([0-9]+?)\.([0-9]+?)" onclick="luV\(([0-9]+?)\)">', r'<vid id="v\1.\2.\3" onclick="luV(\1, \2, \3)">', content)

    # convert UBA bible link
    if '''<ref onclick='document.title="BIBLE:::''' in content:
        def convert_uba_bible_link(match):
            parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
            refs = parser.extractAllReferences(match.group(1))
            ref = parser.bcvToVerseReference(*refs[0])
            if refs:
                return f'''<ref onclick="bcv{refs[0]}">{ref}'''
            return match.group(0)

        content = re.sub(r'''<ref onclick='document.title="BIBLE:::([^<>]+?)"'>\1''', convert_uba_bible_link, content)

    # add notes
    if app.storage.user["notes"]:
        content = re.sub(f">({'|'.join(verses_with_notes)})</vid>", rf'''>\1</vid> <ref onclick="note({b},{c},\1)">📝</ref>''', content)
    # Convert onclick and ondblclick links
    content = content.replace("luV(", "luV1(" if area == 1 else "luV2(")
    content = re.sub(r'''(onclick|ondblclick)="(note|cr|bcv|luV1|luV2|luW|lex|bdbid|etcbcmorph|rmac|searchLexicalEntry|searchWord)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
    content = re.sub(r"""(onclick|ondblclick)='(note|cr|bcv|luV1|luV2|luW|lex|bdbid|etcbcmorph|rmac|searchLexicalEntry|searchWord)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", content)

    # adjust spacing
    content = content.replace("</vid> <pb/><br>", "</vid> ")
    content = content.replace("<br>&emsp;&emsp;", "<br><br>&emsp;&emsp;")
    content = content.replace("<br><br><br>&emsp;&emsp;", "<br><br>&emsp;&emsp;")
    content = content.replace("<br><br>", "<hr>")
    content = content.replace("<br>", "")
    content = content.replace("<hr>", "<br>")

    # Bible Selection menu
    def previous_chapter(selection):
        selected_text, selected_b, selected_c, _ = selection
        bookList = getBibleBookList(db)
        chapterList = getBibleChapterList(db, selected_b)
        if len(chapterList) == 1 or selected_c == chapterList[0]:
            if selected_b == bookList[0]:
                new_b = bookList[-1]
                new_c = getBibleChapterList(db, new_b)[-1]
            else:
                new_b = selected_b - 1
                for i in bookList:
                    previous_book = None
                    if i == selected_b and previous_book is not None:
                        new_b = previous_book
                        break
                    else:
                        previous_book = i
                new_c = getBibleChapterList(db, new_b)[-1]
        else:
            new_b = selected_b
            new_c = selected_c - 1
            for i in chapterList:
                previous_chapter = None
                if i == selected_c and previous_chapter is not None:
                    new_c = previous_chapter
                    break
                else:
                    previous_chapter = i
        if area == 1:
            gui.change_area_1_bible_chapter(selected_text, new_b, new_c, 1)
        else:
            gui.change_area_2_bible_chapter(selected_text, new_b, new_c, 1)

    def next_chapter(selection):
        selected_text, selected_b, selected_c, _ = selection
        bookList = getBibleBookList(db)
        chapterList = getBibleChapterList(db, selected_b)
        if len(chapterList) == 1 or selected_c == chapterList[-1]:
            if selected_b == bookList[-1]:
                new_b = bookList[0]
                new_c = getBibleChapterList(db, new_b)[0]
            else:
                new_b = selected_b + 1
                for i in bookList:
                    previous_book = None
                    if previous_book is not None:
                        new_b = i
                        break
                    elif i == selected_b:
                        previous_book = i
                new_c = getBibleChapterList(db, new_b)[0]
        else:
            new_b = selected_b
            new_c = selected_c + 1
            for i in chapterList:
                previous_chapter = None
                if previous_chapter is not None:
                    new_c = i
                    break
                elif i == selected_c:
                    previous_chapter = i
        if area == 1:
            gui.change_area_1_bible_chapter(selected_text, new_b, new_c, 1)
        else:
            gui.change_area_2_bible_chapter(selected_text, new_b, new_c, 1)

    def additional_items():
        nonlocal gui, bible_selector, area

        def open_tool(selection, title=""):
            app.storage.user['tool_book_text'], app.storage.user['tool_book_number'], app.storage.user['tool_chapter_number'], app.storage.user['tool_verse_number'] = selection
            gui.select_empty_area2_tab()
            gui.load_area_2_content(title=title, sync=False)
        def search_bible(q=""):
            app.storage.user['tool_query'] = q
            gui.select_empty_area2_tab()
            gui.load_area_2_content(title="Verses")
        with ui.button(icon='more_vert').props(f'flat round color={"white" if app.storage.user["dark_mode"] else "black"}'):
            with ui.menu():
                ui.menu_item(f'◀️ {get_translation("Prev Chapter")}', on_click=lambda: previous_chapter(bible_selector.get_selection()))
                ui.menu_item(f'▶️ {get_translation("Next Chapter")}', on_click=lambda: next_chapter(bible_selector.get_selection()))
                if area == 1:
                    ui.separator()
                    ui.menu_item(f'🔍 {get_translation("Search")} {bible_selector.version_select.value}', on_click=lambda: search_bible())
                    ui.menu_item(f'🔍 {get_translation("Search")} {get_translation("OT")}', on_click=lambda: search_bible(q=f"OT:::{app.storage.user['tool_query']}"))
                    ui.menu_item(f'🔍 {get_translation("Search")} {get_translation("NT")}', on_click=lambda: search_bible(q=f"NT:::{app.storage.user['tool_query']}"))
                    ui.menu_item(f'🔍 {get_translation("Search")} {bible_selector.book_select.value}', on_click=lambda: search_bible(q=f"{bible_selector.book_select.value}:::{app.storage.user['tool_query']}"))
                ui.separator()
                ui.menu_item(f'🧩 {get_translation("Book Tools")}', on_click=lambda: book_studies(VerseEventObj(args=bible_selector.get_selection())))
                ui.menu_item(f'🧩 {get_translation("Chapter Tools")}', on_click=lambda: chapter_studies(VerseEventObj(args=bible_selector.get_selection())))
                ui.menu_item(f'🧩 {get_translation("Verse Tools")}', on_click=lambda: luV1_m(VerseEventObj(args=bible_selector.get_selection())) if area == 1 else luV2_m(VerseEventObj(args=bible_selector.get_selection())))
                if config.google_client_id and config.google_client_secret:
                    ui.separator()
                    ui.menu_item(f'📝 {get_translation("Book Note")}', on_click=lambda: open_tool((bible_selector.selected_version, bible_selector.selected_book, 0, 0), title="Notes"))
                    ui.menu_item(f'📝 {get_translation("Chapter Note")}', on_click=lambda: open_tool((bible_selector.selected_version, bible_selector.selected_book, bible_selector.selected_chapter, 0), title="Notes"))
                    ui.menu_item(f'📝 {get_translation("Verse Note")}', on_click=lambda: open_tool(bible_selector.get_selection(), title="Notes"))

    bible_selector.create_ui(title, b, c, v, additional_items=additional_items)

    # create a dummy label for being the parent of `open_verse_context_menu`, as ui.html can't take two context menus
    # without a parent, the context menu doesn't close automatically
    dummy_style = 'position: absolute; width: 0; height: 0; overflow: hidden;'
    if area == 1:
        dummy_label1 = ui.label().style(dummy_style)
    else:
        dummy_label2 = ui.label().style(dummy_style)
    # Render the HTML inside a styled container
    chapter_label = (tab1+"_chapter") if area == 1 else (tab2+"_chapter")
    ui.html(f'<div id="{chapter_label}" class="bible-text">{content}</div>', sanitize=False).classes(f'w-full {chapter_label}')
    # Do not attach a context menu directly to ui.html, which make text unable to be selected on mobile device.

    # chapter buttons
    with ui.row().classes('w-full justify-center pb-[70vh]') as resume_container:
        ui.button(f'◀️ {get_translation("Prev Chapter")}', on_click=lambda: previous_chapter(bible_selector.get_selection()))
        ui.button(f'{get_translation("Next Chapter")} ▶️', on_click=lambda: next_chapter(bible_selector.get_selection()))

    # After the page is built and ready, run our JavaScript
    if (not area == 1) and tab1 and tab2:
        ui.run_javascript(f"""
            {SYNC_JS}
            
            {get_sync_fx(tab1, tab2)}
        """)

    # scrolling, e.g.
    ui.run_javascript(f'scrollToVerse("v{b}.{c}.{v}")')


from biblemateweb import BIBLEMATEWEB_DATA, load_vectors_from_db, get_translation
from nicegui import ui, app, run
from agentmake.utils.rag import get_embeddings, cosine_similarity_matrix
import numpy as np
import re, apsw, os, json, traceback, asyncio
from biblemateweb.data.cr_books import cr_books

def fetch_bible_locations_entry(path):
    db = os.path.join(BIBLEMATEWEB_DATA, "data", "exlb3.data")
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        sql_query = "SELECT content FROM exlbl WHERE path=? limit 1"
        cursor.execute(sql_query, (path,))
        fetch = cursor.fetchone()
        content = fetch[0] if fetch else ""
    vdb = os.path.join(BIBLEMATEWEB_DATA, "data", "exlb3.data")
    location, lat, lng = None, None, None
    with apsw.Connection(vdb) as connn:
        cursor = connn.cursor()
        sql_query = "SELECT location, lat, lng FROM exlbli WHERE path=?"
        cursor.execute(sql_query, (path,))
        if fetch := cursor.fetchone():
            location, lat, lng = fetch
            lat, lng = float(lat), float(lng)
    return content, location, lat, lng

async def fetch_bible_locations_matches_async(query):
    n = ui.notification("Loading ...", timeout=None, spinner=True)
    db_file = os.path.join(BIBLEMATEWEB_DATA, "vectors", "exlb.db")
    sql_table = "exlbl"
    embedding_model = "paraphrase-multilingual"
    path = ""
    options = []

    try:
        # 1. Exact Match Check (Fast, safe on main thread)
        exact_match_found = False
        with apsw.Connection(db_file) as connection:
            cursor = connection.cursor()
            cursor.execute(f"SELECT * FROM {sql_table} WHERE entry = ?;", (query,))
            rows = cursor.fetchall()
            
            if len(rows) == 1:
                path = rows[0][0]
                exact_match_found = True
            elif len(rows) > 1:
                options = [f"[{row[0]}] {row[1]}" for row in rows]
                exact_match_found = True

        # 2. Similarity Search (Offloaded to prevent blocking)
        if not exact_match_found:
            # A. Get Query Vector
            query_vector = await run.io_bound(get_embeddings, [query], embedding_model)
            query_vector = query_vector[0]

            # B. Load & Parse Vectors (Run in separate process)
            entries, document_matrix = await run.cpu_bound(load_vectors_from_db, db_file, sql_table)

            if not entries:
                return []

            # C. Compute Similarity
            similarities = await run.cpu_bound(cosine_similarity_matrix, query_vector, document_matrix)
            
            # D. Sort & Select Top Matches
            top_indices = np.argsort(similarities)[::-1][:app.storage.user["top_similar_entries"]]
            options = [entries[i] for i in top_indices]

    except Exception as ex:
        print("Error during database operation:", ex)
        traceback.print_exc()
        n.message = f'Error: {str(ex)}'
        n.type = 'negative'
        return
    finally:
        n.dismiss()
        
    return path, options

def fetch_bible_locations_matches(query):
    db_file = os.path.join(BIBLEMATEWEB_DATA, "vectors", "exlb.db")
    sql_table = "exlbl"
    embedding_model="paraphrase-multilingual"
    path=""
    options=[]
    try:
        with apsw.Connection(db_file) as connection:
            # search for exact match first
            cursor = connection.cursor()
            cursor.execute(f"SELECT * FROM {sql_table} WHERE entry = ?;", (query,))
            rows = cursor.fetchall()
            if not rows: # perform similarity search if no an exact match
                # convert query to vector
                query_vector = get_embeddings([query], embedding_model)[0]
                # fetch all entries
                cursor.execute(f"SELECT path, entry, entry_vector FROM {sql_table}")
                all_rows = [(f"[{path}] {entry}", entry_vector) for path, entry, entry_vector in cursor.fetchall()]
                if not all_rows:
                    return []
                # build a matrix
                entries, entry_vectors = zip(*[(row[0], np.array(json.loads(row[1]))) for row in all_rows if row[0] and row[1]])
                document_matrix = np.vstack(entry_vectors)
                # perform a similarity search
                similarities = cosine_similarity_matrix(query_vector, document_matrix)
                top_indices = np.argsort(similarities)[::-1][:app.storage.user["top_similar_entries"]]
                # return top matches
                options = [entries[i] for i in top_indices]
            elif len(rows) == 1: # single exact match
                path = rows[0][0]
            else:
                options = [f"[{row[0]}] {row[1]}" for row in rows]
    except Exception as ex:
        print("Error during database operation:", ex)
        traceback.print_exc()
        ui.notify('Error during database operation!', type='negative')
        return
    return path, options

def fetch_all_locations():
    vdb = os.path.join(BIBLEMATEWEB_DATA, "data", "exlb3.data")
    with apsw.Connection(vdb) as connn:
        cursor = connn.cursor()
        sql_query = "SELECT location FROM exlbli" # SELECT lat, lng FROM exlbli WHERE path=?
        cursor.execute(sql_query)
        all_locations = [i[0] for i in cursor.fetchall()]
    return all_locations

def search_bible_locations(gui=None, q='', **_):

    last_entry = q

    def cr(event):
        nonlocal gui
        b, c, v, *_ = event.args
        b = cr_books.get(b, b)
        gui.change_area_1_bible_chapter(None, b, c, v)

    def bcv(event):
        nonlocal gui
        b, c, v, *_ = event.args
        gui.change_area_1_bible_chapter(None, b, c, v)
    
    def website(event):
        url, *_ = event.args
        ui.navigate.to(url, new_tab=True)

    ui.on('bcv', bcv)
    ui.on('cr', cr)
    ui.on('website', website)

    # --- Fuzzy Match Dialog ---
    with ui.dialog() as dialog, ui.card().classes('w-full max-w-md'):
        ui.label("Bible Locations ...").classes('text-xl font-bold text-secondary mb-4')
        ui.label("We couldn't find an exact match. Please select one of these topics:").classes('text-secondary mb-4')
        
        # This container will hold the radio selection dynamically
        selection_container = ui.column().classes('w-full')
        
        with ui.row().classes('w-full justify-end mt-4'):
            ui.button(get_translation("Cancel"), on_click=dialog.close).props('flat color=grey')

    # ----------------------------------------------------------
    # Core: Fetch and Display
    # ----------------------------------------------------------

    async def show_entry(path, keep=True):
        nonlocal content_container, gui, dialog, input_field, last_entry

        n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
        content, location, lat, lng = await run.io_bound(fetch_bible_locations_entry, path)
        n.dismiss()

        # Clear existing rows first
        content_container.clear()

        with content_container:
            # Google Map links
            if "Click HERE for a Live Google Map" in content:
                googleEarthLink = f"""https://earth.google.com/web/@{lat},{lng},654.97002989a,1336.11415664d,35y,0h,76.22542174t,-0r"""
                content = content.replace("Click HERE for a Live Google Map</ref>]", f'''Google Map</ref>] [<ref onclick="website('{googleEarthLink}')">Google Earth</ref>]''')
            # convert links, e.g. <ref onclick="bcv(3,19,26)">
            content = re.sub(r'''(onclick|ondblclick)="(cr|bcv|website)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
            content = re.sub(r"""(onclick|ondblclick)='(cr|bcv|website)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", content)
            # remove images
            content = re.sub('<ref onclick="openHtmlImage.*?</ref>', '', content)
            # remove map
            content = content.replace('<div id="map" style="width:100%;height:500px"></div>', "")
            content = re.sub(r'<script.*?>.*?</script>', '', content, flags=re.DOTALL)
            # add a map
            if lat and lng:
                def restore_center():
                    m.set_center((lat, lng))
                    m.set_zoom(9)
                with ui.element('div').style('position: relative; width: 100%; height: 384px;'):
                    m = ui.leaflet(center=(lat, lng), zoom=9).classes('w-full h-96')
                    with ui.button(on_click=restore_center).props('fab-mini color=primary icon=center_focus_strong').style(
                        'position: absolute; top: 20px; right: 20px; z-index: 1000;'
                    ):
                        ui.tooltip("Restore Center")
                marker = m.marker(latlng=(lat, lng))
                # Wait until marker is ready; wait 0.1 seconds to let the JS layer catch up
                async def bind_marker():
                    # 1. Non-blocking Wait: Checks every 0.1s, but lets other app events happen
                    while marker.id is None:
                        await asyncio.sleep(0.1)
                    # 2. Add a tiny buffer for the JavaScript side to render the marker
                    await asyncio.sleep(0.1)
                    # 3. Bind
                    marker.run_method('bindPopup', f"<b>{location}</b><br>{path}")
                # Fire the safe async function
                ui.timer(0.1, bind_marker, once=True)
            # convert colors for dark mode, e.g. <font color="brown">
            if app.storage.user['dark_mode']:
                content = content.replace('color="brown">', 'color="pink">')
                content = content.replace('color="navy">', 'color="lightskyblue">')
                content = content.replace('<table bgcolor="#BFBFBF"', '<table bgcolor="#424242"')
                content = content.replace('<td bgcolor="#FFFFFF">', '<td bgcolor="#212121">')
                content = content.replace('<tr bgcolor="#FFFFFF">', '<tr bgcolor="#212121">')
                content = content.replace('<tr bgcolor="#DFDFDF">', '<tr bgcolor="#303030">')
            # display
            ui.html(f'<div class="content-text">{content}</div>', sanitize=False).classes('mt-10')

            # Handler
            def search_action(entry):
                nonlocal gui
                """Logic when 'Character' button is clicked"""
                app.storage.user["tool_query"] = entry
                gui.select_empty_area2_tab()
                gui.load_area_2_content(title='Maps')

            with ui.row().classes('w-full justify-center q-my-md'):
                ui.button('Show All Verses', icon='auto_stories', on_click=lambda: gui.show_all_verses(path)) \
                    .props('size=lg rounded color=primary')
                ui.button('Maps', icon='search', on_click=lambda: search_action(path)) \
                    .props('size=lg rounded color=primary')

        # Clear input so user can start typing to filter immediately
        last_entry = path
        input_field.value = ""
        # update tab records
        if keep:
            gui.update_active_area2_tab_records(q=path)

    def handle_up_arrow():
        nonlocal last_entry, input_field
        if not input_field.value.strip():
            input_field.value = last_entry

    async def handle_enter(e, keep=True):
        nonlocal input_field, dialog, selection_container, last_entry
        query = input_field.value.strip()
        if not query:
            return
        elif re.search("BL[0-9]+?$", query):
            await show_entry(query, keep=keep)
            return
        last_entry = query

        input_field.disable()

        try:
            path, options = await fetch_bible_locations_matches_async(query)
        except Exception as e:
            # Handle errors (e.g., network failure)
            ui.notify(f'Error: {e}', type='negative')
        finally:
            # ALWAYS re-enable the input, even if an error occurred above
            input_field.enable()
            # Optional: Refocus the cursor so the user can type the next query immediately
            input_field.run_method('focus')

        if options:
            options = list(set(options))
            def handle_selection(selected_option):
                nonlocal dialog
                if selected_option:
                    dialog.close()
                    path, _ = selected_option.split(" ", 1)
                    ui.timer(0, lambda: show_entry(path[1:-1], keep=keep), once=True)

            selection_container.clear()
            with selection_container:
                # We use a radio button for selection
                radio = ui.radio(options).classes('w-full').props('color=primary')
                ui.button('Show Content', on_click=lambda: handle_selection(radio.value)) \
                    .classes('w-full mt-4 bg-blue-500 text-white shadow-md')    
            dialog.open()
        else:
            await show_entry(path, keep=keep)

    # ==============================================================================
    # 3. UI LAYOUT
    # ==============================================================================
    with ui.row().classes('w-full max-w-3xl mx-auto m-0 py-0 px-4 items-center'):
        input_field = ui.input(
            autocomplete=[], # set empty as initial autocomplete value
            placeholder=f'{get_translation("Search for a bible location")}...'
        ).classes('flex-grow text-lg') \
        .props('outlined dense clearable autofocus enterkeyhint="search"')

        input_field.on('keydown.enter.prevent', handle_enter)
        #input_field.on('update:model-value', filter_verses)
        input_field.on('keydown.up', handle_up_arrow)
        with input_field.add_slot('append'):
            ui.icon('history') \
                .on('click', handle_up_arrow) \
                .classes('text-sm cursor-pointer text-secondary').tooltip('Restore last entry')

        # update autocomplete
        async def get_all_locations():
            all_locations = await run.io_bound(fetch_all_locations)
            input_field.set_autocomplete(all_locations)
        n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
        ui.timer(0, get_all_locations, once=True)
        n.dismiss()

    # --- Main Content Area ---
    with ui.column().classes('w-full items-center'):
        # Define the container HERE within the layout structure
        content_container = ui.column().classes('w-full transition-all !gap-1')

    if q:
        input_field.value = q
        ui.timer(0, lambda: handle_enter(None, keep=False), once=True)
    else:
        input_field.run_method('focus')


from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
from biblemateweb.fx.bible import get_bible_content
from biblemateweb import BIBLEMATEWEB_DATA, config, get_translation, resolve_verses_additional_options
from functools import partial
from nicegui import ui, app, run
import re, apsw, os


def search_bible_verses(gui=None, q='', **_):

    last_entry = q
    default_placeholder = get_translation("Search for words or refs")
    multiple_bibles = None

    def get_bibles():
        nonlocal multiple_bibles, gui
        return multiple_bibles.value if multiple_bibles and multiple_bibles.value else gui.get_area_1_bible_text()

    SQL_QUERY = "PRAGMA case_sensitive_like = false; SELECT Book, Chapter, Verse, Scripture FROM Verses WHERE (Scripture REGEXP ?) ORDER BY Book, Chapter, Verse"

    # --- Data: 66 Bible Books & ID Mapping ---
    BIBLE_BOOKS = [BibleBooks.abbrev["eng"][str(i)][0] for i in range(1,67)]

    # Logic Sets
    OT_BOOKS = BIBLE_BOOKS[:39]
    NT_BOOKS = BIBLE_BOOKS[39:]
    SET_OT = set(OT_BOOKS)
    SET_NT = set(NT_BOOKS)

    # Map abbreviations to Book IDs (1-66)
    BOOK_MAP = {book: i + 1 for i, book in enumerate(BIBLE_BOOKS)}

    # Initialize with full selection state
    default_bible = gui.get_area_1_bible_text()
    client_bibles, initial_bibles, initial_books, q = resolve_verses_additional_options(q, default_bible, app.storage.client["custom"])
    
    state = {'previous': initial_books}

    # ----------------------------------------------------------
    # Helper: Filter Logic
    # ----------------------------------------------------------
    def filter_verses(e=None):
        """
        Filters visibility based on input.
        Iterates over default_slot.children to find rows.
        """
        total_matches = 0
        # Robustly determine the search text
        text = ""
        if e is not None and hasattr(e, 'value'):
            text = e.value 
        else:
            text = input_field.value 
            
        search_term = text if text else ""
        if not app.storage.user['search_case_sensitivity']:
            search_term = search_term.lower()
        
        # Iterate over the actual children of the container
        for row in verses_container.default_slot.children:
            # Skip elements that aren't our verse rows (if any)
            if not hasattr(row, 'verse_data'):
                continue

            # Explicitly show all if search is empty
            if not search_term:
                row.set_visibility(True)
                continue

            data = row.verse_data
            ref_text = data['ref']
            if not app.storage.user['search_case_sensitivity']:
                ref_text = ref_text.lower()
            clean_content = re.sub('<[^<]+?>', '', data['content'])
            if not app.storage.user['search_case_sensitivity']:
                clean_content = clean_content.lower()

            if not app.storage.user['search_mode'] == 2:
                is_match = (re.search(search_term, ref_text)) or (re.search(search_term, clean_content))
            else:
                is_match = (search_term in ref_text) or (search_term in clean_content)
            row.set_visibility(is_match)
            if is_match:
                total_matches += 1
        if total_matches:
            ui.notify(f"{total_matches} {'match' if total_matches == 1 else 'matches'} found!")

    # ----------------------------------------------------------
    # Helper: Remove Verse
    # ----------------------------------------------------------
    def remove_verse_row(row_element, reference):
        try:
            verses_container.remove(row_element)
            ui.notify(f'Removed: {reference}', type='warning', position='top')
        except Exception as e:
            print(f"Error removing row: {e}")

    # ----------------------------------------------------------
    # Helper: Open Chapter
    # ----------------------------------------------------------
    def open_chapter_next_area2_tab(bible, b, c, v):
        gui.select_next_area2_tab()
        gui.change_area_2_bible_chapter(bible, b, c, v, sync=False)

    def open_chapter_empty_area2_tab(bible, b, c, v):
        gui.select_empty_area2_tab()
        gui.change_area_2_bible_chapter(bible, b, c, v, sync=False)

    # ----------------------------------------------------------
    # Core: Fetch and Display
    # ----------------------------------------------------------
    def handle_up_arrow():
        nonlocal last_entry, input_field
        if not input_field.value.strip():
            input_field.value = last_entry

    async def handle_enter(e, keep=True):
        nonlocal SQL_QUERY, last_entry
        query = input_field.value.strip()
    
        # Clear existing rows first
        verses_container.clear()

        if not query:
            ui.notify('Display cleared', type='positive', position='top')
            input_field.props(f'placeholder="{default_placeholder}"')
            return

        # update tab record
        last_entry = query
        if keep:
            gui.update_active_area2_tab_records(q=query)

        input_field.disable()
        parser = BibleVerseParser(False, language=app.storage.user['ui_language'])

        try:
            n = ui.notification("Loading ...", timeout=None, spinner=True)
            highlights = False
            if search_morphology := re.search(r"^([EG][0-9]+?)\|", query): # search morphology
                multiple_bibles.value = [app.storage.user["tool_book_text"]]
                lexical_entry = search_morphology.group(1) + ",%"
                suffix = ""
                for i in query.split("|")[1:]:
                    suffix += f"AND Morphology LIKE '{i}' "
                if books := re.search("WHERE (Book.*?) AND", SQL_QUERY):
                    suffix += f"AND {books.group(1)} " # limit morphology search in particular books
                db = os.path.join(BIBLEMATEWEB_DATA, "morphology.sqlite")
                with apsw.Connection(db) as connn:
                    query = f"SELECT Book, Chapter, Verse FROM morphology WHERE LexicalEntry LIKE ? {suffix}ORDER BY Book, Chapter, Verse"
                    cursor = connn.cursor()
                    cursor.execute(query, (lexical_entry,))
                    fetch = cursor.fetchall()
                if not fetch:
                    ui.notify('No verses found.', type='negative')
                    return
                verses = await run.io_bound(get_bible_content, bible=get_bibles(), sql_query=SQL_QUERY, refs=fetch, parser=parser)
            elif re.search("^BP[0-9]+?$", query): # bible characters entries
                db_file = os.path.join(BIBLEMATEWEB_DATA, "data", "biblePeople.data")
                with apsw.Connection(db_file) as connn:
                    cursor = connn.cursor()
                    cursor.execute("SELECT Book, Chapter, Verse FROM PEOPLE WHERE PersonID=? ORDER BY Book, Chapter, Verse", (int(query[2:]),))
                    fetch = cursor.fetchall()
                if not fetch:
                    ui.notify('No verses found.', type='negative')
                    return
                verses = await run.io_bound(get_bible_content, bible=get_bibles(), sql_query=SQL_QUERY, refs=fetch, parser=parser)
            elif re.search("^BL[0-9]+?$", query): # bible locatios entries
                db_file = os.path.join(BIBLEMATEWEB_DATA, "indexes2.sqlite")
                with apsw.Connection(db_file) as connn:
                    cursor = connn.cursor()
                    cursor.execute('''SELECT Book, Chapter, Verse FROM exlbl WHERE Information LIKE ? ORDER BY Book, Chapter, Verse''', (f"%'{query}'%",))
                    fetch = cursor.fetchall()
                if not fetch:
                    ui.notify('No verses found.', type='negative')
                    return
                verses = await run.io_bound(get_bible_content, bible=get_bibles(), sql_query=SQL_QUERY, refs=fetch, parser=parser)
            elif re.search(f"^({"|".join(list(config.topics.keys()))})[0-9]+?$", query): # bible topics entries
                db_file = os.path.join(BIBLEMATEWEB_DATA, "indexes2.sqlite")
                with apsw.Connection(db_file) as connn:
                    cursor = connn.cursor()
                    cursor.execute('''SELECT Book, Chapter, Verse FROM exlbt WHERE Information LIKE ? ORDER BY Book, Chapter, Verse''', (f"%'{query}'%",))
                    fetch = cursor.fetchall()
                if not fetch:
                    ui.notify('No verses found.', type='negative')
                    return
                verses = await run.io_bound(get_bible_content, bible=get_bibles(), sql_query=SQL_QUERY, refs=fetch, parser=parser)
            elif re.search(f"^({"|".join(list(config.dictionaries.keys()))})[0-9]+?$", query): # bible dictionaries entries
                db_file = os.path.join(BIBLEMATEWEB_DATA, "indexes2.sqlite")
                with apsw.Connection(db_file) as connn:
                    cursor = connn.cursor()
                    cursor.execute('''SELECT Book, Chapter, Verse FROM dictionaries WHERE Information LIKE ? ORDER BY Book, Chapter, Verse''', (f"%'{query}'%",))
                    fetch = cursor.fetchall()
                if not fetch:
                    ui.notify('No verses found.', type='negative')
                    return
                verses = await run.io_bound(get_bible_content, bible=get_bibles(), sql_query=SQL_QUERY, refs=fetch, parser=parser)
            elif re.search(f"^(ISBE|{"|".join(list(config.encyclopedias.keys()))})[0-9]+?$", query): # bible encyclopedia entries
                db_file = os.path.join(BIBLEMATEWEB_DATA, "indexes2.sqlite")
                with apsw.Connection(db_file) as connn:
                    cursor = connn.cursor()
                    cursor.execute('''SELECT Book, Chapter, Verse FROM encyclopedia WHERE Information LIKE ? ORDER BY Book, Chapter, Verse''', (f"%'{query}'%",))
                    fetch = cursor.fetchall()
                if not fetch:
                    ui.notify('No verses found.', type='negative')
                    return
                verses = await run.io_bound(get_bible_content, bible=get_bibles(), sql_query=SQL_QUERY, refs=fetch, parser=parser)
            else: # regular search
                verses = await run.io_bound(get_bible_content, user_input=query, bible=get_bibles(), sql_query=SQL_QUERY, search_mode=app.storage.user['search_mode'], top_similar_verses=app.storage.user['top_similar_verses'], search_case_sensitivity=app.storage.user['search_case_sensitivity'], parser=parser)
                highlights = True

            if not verses:
                ui.notify('No verses found.', type='negative')
                return

            if not verses[-1]['ref']:
                ui.notify(verses[-1]['content'])
                verses = verses[:-1]
            with verses_container:
                for v in verses:
                    # Row setup
                    with ui.column().classes('w-full shadow-sm rounded-lg items-start no-wrap border border-gray-200 !gap-0') as row:
                        
                        row.verse_data = v # Store data for filter function

                        def get_verse_content(ref, content):
                            content = re.sub("<[^<>]+?>", "", content)
                            return f"[{ref}] {content}"
                        
                        def ask_biblemate(ref, content):
                            nonlocal gui
                            gui.ask_biblemate(f"[{ref}] {re.sub("<[^<>]+?>", "", content)}")

                        # --- Chip (Clickable & Removable) ---
                        with ui.element('div').classes('flex-none pt-1'): 
                            with ui.chip(
                                v['ref'], 
                                removable=True, 
                                icon='book',
                                #on_click=partial(ui.notify, f'Clicked {v['ref']}'),
                            ).props('text-color=white').classes('cursor-pointer font-bold shadow-sm') as chip:
                                with ui.menu():
                                    ui.menu_item('📋 Copy', on_click=partial(gui.copy_text, get_verse_content(v['ref'], v['content'])))
                                    ui.separator()
                                    ui.menu_item('📜 Open in Bible Area', on_click=partial(gui.change_area_1_bible_chapter, v['bible'], v['b'], v['c'], v['v']))
                                    ui.menu_item('📜 Open Here', on_click=partial(gui.change_area_2_bible_chapter, v['bible'], v['b'], v['c'], v['v'], sync=False))
                                    ui.menu_item('📜 Open in Next Tab', on_click=partial(open_chapter_next_area2_tab, v['bible'], v['b'], v['c'], v['v']))
                                    ui.menu_item('📜 Open in New Tab', on_click=partial(open_chapter_empty_area2_tab, v['bible'], v['b'], v['c'], v['v']))
                                    ui.separator()
                                    ui.menu_item('💬 Ask BibleMate', on_click=partial(ask_biblemate, v['ref'], v['content']))
                            chip.on('remove', lambda _, r=row, ref=v['ref']: remove_verse_row(r, ref))

                        # --- Content ---
                        content = v['content']
                        # add tooltip
                        if "</heb>" in content:
                            content = re.sub('(<heb id=")(.*?)"', r'\1\2" data-word="\2" class="tooltip-word"', content)
                            content = content.replace("<heb> </heb>", "<heb>&nbsp;</heb>")
                            content = f"<div style='display: inline-block; direction: rtl;'>{content}</div>"
                        elif "</grk>" in content:
                            content = re.sub('(<grk id=")(.*?)"', r'\1\2" data-word="\2" class="tooltip-word"', content)
                        if highlights:
                            if app.storage.user["dark_mode"]:
                                content = re.sub(f"({query})", r"<font color='orange'>\1</font>", content, flags=0 if app.storage.user['search_case_sensitivity'] else re.IGNORECASE)
                            else:
                                content = re.sub(f"({query})", r"<span style='background-color: orange;'>\1</span>", content, flags=0 if app.storage.user['search_case_sensitivity'] else re.IGNORECASE)
                        ui.html(content, sanitize=False).classes('grow min-w-0 leading-relaxed pl-2 text-base break-words')

            # Clear input so user can start typing to filter immediately
            input_field.value = ""
            input_field.props(f'''placeholder="{get_translation('Type to filter')} {len(verses)} {get_translation('results')}..."''')
            ui.notify(f"{len(verses)} {get_translation('result') if not verses or len(verses) == 1 else get_translation('results')}")

        except Exception as e:
            # Handle errors (e.g., network failure)
            #ui.notify(f'Error: {e}', type='negative')
            n.message = f'Error: {str(e)}'
            n.type = 'negative'
        finally:
            n.dismiss()
            # ALWAYS re-enable the input, even if an error occurred above
            input_field.enable()
            # Optional: Refocus the cursor so the user can type the next query immediately
            input_field.run_method('focus')

    # ==============================================================================
    # 3. UI LAYOUT
    # ==============================================================================
    
    with ui.row().classes('w-full max-w-3xl mx-auto m-0 py-0 px-4 items-center'):
        input_field = ui.input(
            autocomplete=BIBLE_BOOKS,
            placeholder=default_placeholder
        ).classes('flex-grow text-lg') \
        .props('outlined clearable autofocus enterkeyhint="search"')

        input_field.on('keydown.enter.prevent', handle_enter)
        input_field.on('update:model-value', filter_verses)
        input_field.on('keydown.up', handle_up_arrow)
        with input_field.add_slot('append'):
            ui.icon('history') \
                .on('click', handle_up_arrow) \
                .classes('text-sm cursor-pointer text-secondary').tooltip('Restore last entry')

        # 2. Scope Dropdown
        # Options: All, None, OT, NT, then the books
        options = ['All', 'None', 'OT', 'NT'] + BIBLE_BOOKS
        
        scope_select = ui.select(
            options=options,
            #label='Search',
            multiple=True,
            with_input=True
        ).classes('w-22').props('dense clearable')

        def update_sql_query(selected_values):
            """Generates the SQLite query based on selection."""
            nonlocal SQL_QUERY
            
            # Filter to keep ONLY the actual book strings (ignore All/None/OT/NT)
            real_books = [b for b in selected_values if b in BIBLE_BOOKS]
            book_ids = [str(BOOK_MAP[b]) for b in real_books]
            
            base_query = "PRAGMA case_sensitive_like = false; SELECT * FROM Verses"
            where_clauses = []

            # Handle Book Logic
            if 'All' in selected_values:
                pass 
            elif not real_books:
                where_clauses.append("1=0")
            elif len(real_books) == 1:
                where_clauses.append(f"Book={book_ids[0]}")
            else:
                # Optimization: check if it's exactly OT or NT for cleaner SQL?
                # (Optional, but strictly sticking to IDs is safer for the engine)
                where_clauses.append(f"Book IN ({', '.join(book_ids)})")

            where_clauses.append(f"(Scripture REGEXP ?) ORDER BY Book, Chapter, Verse")

            # Assemble
            full_query = base_query
            if where_clauses:
                full_query += " WHERE " + " AND ".join(where_clauses)
            
            #ui.notify(full_query)
            SQL_QUERY = full_query

        def handle_scope_change(e):
            """
            Handles complex logic for All, None, OT, NT and individual books.
            """
            current = e.value if e.value else []
            previous = state['previous']
            
            # 1. Determine Triggers
            added = set(current) - set(previous)
            removed = set(previous) - set(current)
            
            # Start with the currently selected actual books
            selected_books = set(x for x in current if x in BIBLE_BOOKS)

            # 2. Apply High-Level Triggers
            # Priority: None > All > OT/NT > Individual removals

            if 'None' in added:
                selected_books.clear()
            
            elif 'All' in added:
                selected_books = set(BIBLE_BOOKS)
            
            elif 'OT' in added:
                selected_books.update(SET_OT)
            
            elif 'NT' in added:
                selected_books.update(SET_NT)
            
            # Handle Removals of Groups
            # We only remove the group's books if the group TAG was explicitly removed
            elif 'All' in removed and len(removed) == 1:
                 # User clicked 'All' to uncheck it -> Clear all
                 selected_books.clear()
            
            elif 'OT' in removed:
                # Check if OT tag was explicitly removed (not just because a child book was clicked)
                # If a child was clicked, 'removed' contains {'ChildBook'}.
                # If OT tag was clicked, 'removed' contains {'OT'}.
                # Note: NiceGUI might remove OT from 'current' automatically if child removed,
                # but 'removed' set captures the diff.
                if not (removed & SET_OT): # If no individual OT books were in the removed set
                    selected_books -= SET_OT

            elif 'NT' in removed:
                if not (removed & SET_NT):
                    selected_books -= SET_NT

            # 3. Reconstruct Selection State
            # We rebuild the list from scratch based on the books we decided are selected
            new_selection = []
            
            # Helper: Check completeness
            has_ot = SET_OT.issubset(selected_books)
            has_nt = SET_NT.issubset(selected_books)
            has_all = len(selected_books) == 66
            is_empty = len(selected_books) == 0

            # Add Meta Tags
            if has_all:
                new_selection.append('All')
            if is_empty:
                new_selection.append('None')
            if has_ot:
                new_selection.append('OT')
            if has_nt:
                new_selection.append('NT')

            # Add Books (maintain order)
            for book in BIBLE_BOOKS:
                if book in selected_books:
                    new_selection.append(book)

            # 4. Update UI and State
            if set(new_selection) != set(current):
                scope_select.value = new_selection
                state['previous'] = new_selection
                update_sql_query(new_selection)
                return

            state['previous'] = current
            update_sql_query(current)

        # Checkbox
        ui.checkbox(
            get_translation('Case-sensitive')
        ).bind_value(
            app.storage.user, 'search_case_sensitivity'
        ).props('dense')
            
        # Radio Buttons
        # options: dictionary maps the stored value (keys) to the display label (values)
        # props('inline'): makes the radio buttons layout horizontally
        modes = ui.radio(
            options={1: get_translation('Literal'), 2: get_translation('Regex'), 3: get_translation('Semantic')},
        ).bind_value(
            app.storage.user, 'search_mode'
        ).props('dense inline color=primary')
        #modes.tooltip = ui.tooltip('Search Modes:\n1. Literal search for plain text\n2. Search for regular expression\n3. Semantic search for meaning')

        # Multi-select dropdown
        multiple_bibles = ui.select(
            client_bibles,
            value=initial_bibles,
            #label='Select Bibles to search', 
            multiple=True,
            with_input=True,
        ).classes('grow').props('use-chips outlined dense clearable')

        # --- Bindings ---
        scope_select.on_value_change(handle_scope_change)

        # Initialize
        scope_select.value = initial_books

    # --- Main Content Area ---
    with ui.column().classes('w-full items-center'):
        # Define the container HERE within the layout structure
        verses_container = ui.column().classes('w-full transition-all !gap-1')

    if q:
        input_field.value = q
        #handle_enter(None, keep=False) # RuntimeWarning: coroutine 'search_bible_verses.<locals>.handle_enter' was never awaited
        ui.timer(0, lambda: handle_enter(None, keep=False), once=True)
    else:
        input_field.run_method('focus')


from biblemateweb import get_translation
from biblemateweb.data.bible_locations import BIBLE_LOCATIONS
from biblemateweb.fx.location_finder import LocationFinder
from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
from nicegui import ui, app
from functools import partial
import math, re, asyncio


# --- Data: 66 Bible Books & ID Mapping ---
BIBLE_BOOKS = [BibleBooks.abbrev["eng"][str(i)][0] for i in range(1,67)]

# Create a dictionary for Dropdown options: {ID: "Name (ID)"}
# This handles duplicate names by ensuring the value passed is the unique ID
LOCATION_OPTIONS = {
    uid: data[0] for uid, data in BIBLE_LOCATIONS.items()
}

# --- 2. HELPER FUNCTIONS ---

def haversine_distance(coord1, coord2, unit='km'):
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    """
    lat1, lon1 = coord1
    lat2, lon2 = coord2
    
    # Convert decimal degrees to radians 
    lat1, lon1, lat2, lon2 = map(math.radians, [lat1, lon1, lat2, lon2])

    # Haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = math.sin(dlat/2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon/2)**2
    c = 2 * math.asin(math.sqrt(a)) 
    r = 6371 # Radius of earth in kilometers
    
    distance = c * r
    
    if unit == 'miles':
        return distance * 0.621371
    return distance

# --- 3. UI LAYOUT ---

def search_bible_maps(gui=None, q='', **_):

    last_entry = q

    def handle_up_arrow():
        nonlocal last_entry, search_input
        if not search_input.value.strip():
            search_input.value = last_entry

    def exlbl(event):
        nonlocal gui
        app.storage.user['tool_query'], *_ = event.args
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Locations')
    ui.on('exlbl', exlbl)

    jerusalem = (31.777, 35.235)
    parser = BibleVerseParser(False)
    finder = LocationFinder()

    location_multiselect = None
    # Apply a full height column with no wrap so the map can stretch
    with ui.column().classes('w-full h-screen no-wrap p-4 gap-4'):

        # Dictionary to keep track of added layers {loc_id: layer}
        # Leaflet in NiceGUI doesn't have a direct "get_marker_by_id", so we track them locally
        active_markers = {} 

        def update_map_markers(selected_ids):
            """
            Synchronizes the map markers with the list of selected IDs.
            """
            current_ids = set(active_markers.keys())
            target_ids = set(selected_ids)

            # 1. Remove markers that are no longer selected
            to_remove = current_ids - target_ids
            for uid in to_remove:
                bible_map.remove_layer(active_markers[uid])
                del active_markers[uid]

            # 2. Add new markers
            to_add = target_ids - current_ids
            for uid in to_add:
                name, lat, lon = BIBLE_LOCATIONS[uid]
                # Add marker with popup
                marker = bible_map.marker(latlng=(lat, lon))
                # The following line does not work when tool_query parameter is loaded when the page is opened
                #marker.run_method('bindPopup', f'''<b>{name}</b><br>[<ref onclick="emitEvent('exlbl', ['{uid}']); return false;">{uid}</ref>]''')
                # Wait until marker is ready; wait 0.1 seconds to let the JS layer catch up
                async def bind_marker(marker, name, uid):
                    # 1. Non-blocking Wait: Checks every 0.1s, but lets other app events happen
                    while marker.id is None:
                        await asyncio.sleep(0.1)
                    # 2. Add a tiny buffer for the JavaScript side to render the marker
                    await asyncio.sleep(0.5)
                    # 3. Bind
                    marker.run_method('bindPopup', f'''<b>{name}</b><br>[<ref onclick="emitEvent('exlbl', ['{uid}']); return false;">{uid}</ref>]''')
                # Fire the safe async function
                ui.timer(0.1, partial(bind_marker, marker, name, uid), once=True)
                
                active_markers[uid] = marker
                
            # If we added exactly one new marker, pan to it
            if len(to_add) == 1:
                uid = list(to_add)[0]
                lat, lon = BIBLE_LOCATIONS[uid][1], BIBLE_LOCATIONS[uid][2]
                bible_map.set_center((lat, lon))
                bible_map.set_zoom(9)
            else:
                fit_all_markers()


        # ==========================================
        # CONTROLS
        # ==========================================
        with ui.card().classes('w-full'):
            #ui.label('📍 Map Explorer').classes('text-sm font-bold text-gray-500')
            
            with ui.row().classes('w-full items-center gap-4 !p-0'):
                
                # Prepare options for multiselect with "All" and "None"
                # Using special keys that we can intercept
                multi_options = {
                    'CMD_ALL': 'All', 
                    'CMD_NONE': 'None'
                }
                multi_options.update(LOCATION_OPTIONS)

                # Multi-select dropdown
                location_multiselect = ui.select(
                    multi_options, 
                    label=get_translation('Locations'), 
                    multiple=True,
                    with_input=True
                ).classes('w-30').props('dense clearable')  # min-w-[40px]

                # Text Input for quick search
                search_input = ui.input(
                    #label='Search',
                    autocomplete=list(LOCATION_OPTIONS.values())+BIBLE_BOOKS,
                    placeholder=f'{get_translation("Search for location(s) or verse reference(s)")}...',
                ).classes('flex-grow text-lg').props('outlined clearable autofocus enterkeyhint="search"')
                
                def on_search_enter(keep=True):
                    """Finds a location by name and adds it to the multiselect (which triggers map update)"""
                    nonlocal search_input, location_multiselect, last_entry, gui

                    query = search_input.value.strip()
                    if not query: return
                    last_entry = query

                    if keep:
                        gui.update_active_area2_tab_records(q=query)

                    current_vals = location_multiselect.value or []

                    # when users enter verse reference(s)
                    if verseList := parser.extractAllReferences(query, tagged=False):
                        combinedLocations = []
                        for reference in verseList:
                            combinedLocations += finder.getLocationsFromReference(reference)
                        if found_id := sorted(list(set(combinedLocations))):
                            location_multiselect.value = list(set(current_vals + found_id))
                            ui.notify(f"{len(found_id)} locations found!")
                            search_input.value = ""
                            return
                    
                    # when users enter location id(s)
                    if re.search("^BL[0-9BL, ]+?$", query):
                        found_id = [i.strip() for i in query.split(",") if i.strip() in BIBLE_LOCATIONS]
                        if found_id:
                            location_multiselect.value = list(set(current_vals + found_id))
                            ui.notify(f"{len(found_id)} locations found!")
                            search_input.value = ""
                            if len(found_id) == 2:
                                loc1_select.value, loc2_select.value = found_id
                                calculate()
                            return

                    # when users enter location name(s)
                    query = search_input.value.lower()

                    found_id = []
                    for i in query.split(","):
                        i = i.strip()
                        if not i: continue
                        for uid, data in BIBLE_LOCATIONS.items():
                            if i in data[0].lower():
                                found_id.append(uid)
                                break
                    
                    if found_id:
                        # This update will trigger the on_value_change event
                        location_multiselect.value = list(set(current_vals + found_id))
                        ui.notify(f"{len(found_id)} locations found!")
                        search_input.value = "" # clear input
                    else:
                        ui.notify("Location not found", type='warning')

                search_input.on('keydown.enter.prevent', on_search_enter)
                search_input.on('keydown.up', handle_up_arrow)
                with search_input.add_slot('append'):
                    ui.icon('history') \
                        .on('click', handle_up_arrow) \
                        .classes('text-sm cursor-pointer text-secondary').tooltip('Restore last entry')

                # Intercept selection to handle "All" and "None" logic
                def handle_selection_change(e):
                    selected_values = e.value
                    
                    # Handle "All"
                    if 'CMD_ALL' in selected_values:
                        all_real_ids = list(LOCATION_OPTIONS.keys())
                        location_multiselect.value = all_real_ids
                        update_map_markers(all_real_ids)
                        return

                    # Handle "None"
                    if 'CMD_NONE' in selected_values:
                        location_multiselect.value = []
                        update_map_markers([])
                        return

                    # Normal update
                    update_map_markers(selected_values)

                # Bind the custom handler
                location_multiselect.on_value_change(handle_selection_change)

        # ==========================================
        # DISTANCE CALCULATOR
        # ==========================================
        with ui.card().classes('w-full'):
            #ui.label('📏 Bible Location Distance Calculator').classes('text-lg font-bold text-gray-700 mb-2')
            
            with ui.row().classes('w-full items-center gap-4 p-0'):
                # Location Selectors
                loc1_select = ui.select(LOCATION_OPTIONS, label=get_translation('From'), with_input=True).classes('w-35').props('dense')
                loc2_select = ui.select(LOCATION_OPTIONS, label=get_translation('To'), with_input=True).classes('w-35').props('dense')
                
                # Unit Toggle
                unit_radio = ui.radio({'km': get_translation('km'), 'miles': get_translation('miles')}, value='km').props('dense inline')
                
                # Result Label
                result_label = ui.label(get_translation('Distance')).classes('dense text-lg font-medium text-secondary ml-auto mr-4')

                # Calculation Logic
                def calculate():
                    #nonlocal location_multiselect, result_label
                    if loc1_select.value and not loc1_select.value in location_multiselect.value:
                        location_multiselect.value = location_multiselect.value + [loc1_select.value]
                    if loc2_select.value and not loc2_select.value in location_multiselect.value:
                        location_multiselect.value = location_multiselect.value + [loc2_select.value]
                    if (loc1_select.value and not loc2_select.value) or (not loc1_select.value and loc2_select.value):
                        result_label.text = "Select one more"
                        return
                    
                    # Get coordinates from ID
                    id1 = loc1_select.value
                    id2 = loc2_select.value
                    
                    coord1 = (BIBLE_LOCATIONS[id1][1], BIBLE_LOCATIONS[id1][2])
                    coord2 = (BIBLE_LOCATIONS[id2][1], BIBLE_LOCATIONS[id2][2])
                    
                    dist = haversine_distance(coord1, coord2, unit_radio.value)
                    unit_label = get_translation('km') if unit_radio.value == 'km' else get_translation('miles')
                    
                    result_label.text = f"{dist:.2f} {unit_label}"

                # Trigger calculation on button click or change
                #ui.button('Calculate', on_click=calculate).classes('bg-blue-600 text-white')
                
                # Auto-calculate when inputs change
                loc1_select.on_value_change(calculate)
                loc2_select.on_value_change(calculate)
                unit_radio.on_value_change(calculate)

        # ==========================================
        # LEAFLET MAP
        # ==========================================

        # Fit bounds using JavaScript
        def fit_all_markers():
            if not location_multiselect.value:
                bible_map.set_center(jerusalem) # Jerusalem
                bible_map.set_zoom(9)
                return
            locations = [BIBLE_LOCATIONS[uid][1:] for uid in location_multiselect.value]
            lats = [loc[0] for loc in locations]
            lngs = [loc[1] for loc in locations]
            #center = (sum(lats)/len(lats), sum(lngs)/len(lngs))
            bounds = [[min(lats), min(lngs)], [max(lats), max(lngs)]]
            ui.run_javascript(f'''
                setTimeout(() => {{
                    const element = getElement({bible_map.id});
                    if (element && element.map) {{
                        const bounds = L.latLngBounds({bounds});
                        element.map.fitBounds(bounds, {{
                            padding: [50, 50],
                            maxZoom: 12
                        }});
                    }}
                }}, 100);
            ''')

        with ui.element('div').style('position: relative; width: 100%; display: flex; flex-direction: column; height: 100%;'):
            bible_map = ui.leaflet(center=jerusalem, zoom=9).classes('w-full flex-grow rounded-lg shadow-md') # default; center on Jerusalem approx
            with ui.button(on_click=fit_all_markers).props('fab-mini color=primary icon=center_focus_strong').style(
                'position: absolute; top: 20px; right: 20px; z-index: 1000;'
            ):
                ui.tooltip('Zoom to Fit')

        if q:
            search_input.value = q
            on_search_enter(keep=False)


import re, zipfile, json, io, markdown2, re, base64
import urllib.parse
from nicegui import ui, app, run
from biblemateweb import get_translation, loading
from biblemateweb.fx.bible import BibleSelector
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
from biblemateweb.fx.cloud_index_manager import get_drive_service, CloudIndexManager
from googleapiclient.http import MediaIoBaseUpload


class CloudNotepad:
    def __init__(self, content=""):
        self.text_content = content
        self.is_editing = True
        self.parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
        
    def setup_ui(self):

        # --- Content Area ---
        # Card must be 'flex flex-col' so the child (textarea) can grow
        with ui.card().classes('w-full h-[75vh] p-0 flex flex-col'):
            
            # 1. Edit Mode: Text Area
            # We apply our custom 'full-height-textarea' class here
            self.textarea = ui.textarea(
                placeholder=get_translation("Start typing your notes here..."),
                value=self.text_content
            ).classes('w-full flex-grow full-height-textarea p-2 border-none focus:outline-none') \
             .props('flat squares resize-none') \
             .bind_visibility_from(self, 'is_editing')

            # 2. Read Mode: HTML Preview
            with ui.scroll_area().classes('w-full flex-grow p-2') \
                    .bind_visibility_from(self, 'is_editing', backward=lambda x: not x):
                self.html_view = ui.html(f'<div class="content-text">{self.text_content}</div>', sanitize=False).classes('w-full prose max-w-none')

    def toggle_mode(self):
        self.is_editing = not self.is_editing
        if not self.is_editing:
            self.update_preview()

    def update_preview(self):
        content = self.textarea.value or ''
        try:
            # Added your requested extras
            html_content = markdown2.markdown(
                content, 
                extras=["tables", "fenced-code-blocks", "toc", "cuddled-lists"]
            )
            html_content = self.parser.parseText(html_content)
            html_content = re.sub(r'''(onclick|ondblclick)="(bdbid|lex|cr|bcv|website)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', html_content)
            html_content = re.sub(r"""(onclick|ondblclick)='(bdbid|lex|cr|bcv|website)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", html_content)
            self.html_view.content = html_content
        except Exception as e:
            self.html_view.content = f"<p class='text-red-500'>Error: {str(e)}</p>"

    def download_file(self):
        content = self.textarea.value or ''
        if not content:
            ui.notify('Nothing to download!', type='warning')
            return
        ui.download(content.encode('utf-8'), 'biblemate_notes.md')
        ui.notify('Downloaded!', type='positive')
    
    async def handle_upload(self, e):
        try:
            content = await e.file.read()
            self.textarea.value = content.decode('utf-8')
            if not self.is_editing: self.update_preview()
            ui.notify('Loaded!', type='positive')
            self.upload.reset()
        except Exception as ex:
            ui.notify(f'Error: {str(ex)}', type='negative')
    
    def clear_text(self):
        self.textarea.value = ''
        self.html_view.content = ''
        ui.notify('Cleared!', type='info')

def notes(gui=None, bt=None, b=1, c=1, v=1, area=2, **_):
    if not bt:
        bt = gui.get_area_1_bible_text()
    elif bt in ("ORB", "OIB", "OPB", "ODB", "OLB"):
        bt = "OHGB"
    
    # Book Note
    if c == 0:
        v = 0

    # 1. Auth Check
    token = app.storage.user.get('google_token', "")
    if not token:
        with ui.card().classes('absolute-center'):
            ui.html('Sign in with Google to securely save and sync your personal notes across all your devices.<br><i><b>Data Policy Note:</b> BibleMate AI does not collect or store your personal notes. Your notes are saved directly within your own Google Account.</i>', sanitize=False)
            with ui.row().classes('w-full justify-center'):
                ui.button('Login with Google', on_click=lambda: ui.navigate.to('/login'))
            with ui.expansion(get_translation("BibleMate AI Data Policy & Privacy Commitment"), icon='privacy_tip').props('header-class="text-secondary"'):
                ui.html(
                    """<b>We respect your privacy.</b>
To keep your data private, BibleMate AI doesn't collect, store or share your personal information on its servers. 
Instead, daily logs or bible notes are saved exclusively on <b>your own Google Drive/Account.</b>, ensuring that you retain full control and ownership of your private data at all times.""", 
                    sanitize=False,
                )
        return

    bible_selector = None
    service = get_drive_service(token)
    index_mgr = CloudIndexManager(service)
    notepad = CloudNotepad()

    def change_note(version=None, book=1, chapter=1, verse=1):
        nonlocal gui
        _, app.storage.user['tool_book_number'], app.storage.user['tool_chapter_number'], app.storage.user['tool_verse_number'] = version, book, chapter, verse
        gui.load_area_2_content(title='Notes', sync=False)
    bible_selector = BibleSelector(version_options=[bt], on_book_changed=change_note, on_chapter_changed=change_note, on_verse_changed=change_note, chapter_zero=True, verse_zero=True)

    def refresh_ui():
        try:
            nonlocal gui, b, c, v
            active_area1_tab = gui.get_active_area1_tab()
            if active_area1_tab in app.storage.user and app.storage.user[active_area1_tab]["b"] == b and app.storage.user[active_area1_tab]["c"] == c:
                gui.change_area_1_bible_chapter(book=b, chapter=c, verse=v)
        except:
            import traceback
            traceback.print_exc()

    def get_filename(verse_id):
        return f"{verse_id}.json"

    def get_vid(): 
        return f"{b}_{c}_{v}"

    def load_current_note():
        vid = get_vid()
        try:
            filename = get_filename(vid)
            results = service.files().list(
                q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                spaces='appDataFolder',
                fields='files(id)'
            ).execute()
            files = results.get('files', [])
            
            if files:
                request = service.files().get_media(fileId=files[0]['id'])
                data = json.loads(request.execute())
                return data.get("content", "")
            return ""
        except Exception as e:
            ui.notify(f"Error loading: {e}", type='negative')
            return ""

    def save_note_sync(service, verse_id, content, filename):
        """
        Performs the slow network calls to Google Drive.
        Running this on the main thread would freeze the app.
        """
        import json, io
        from googleapiclient.http import MediaIoBaseUpload
        
        # Prepare the file data
        file_data = {"verse_id": verse_id, "content": content}
        media = MediaIoBaseUpload(
            io.BytesIO(json.dumps(file_data).encode('utf-8')), 
            mimetype='application/json'
        )
        
        # 1. Search for existing file
        results = service.files().list(
            q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
            spaces='appDataFolder',
            fields='files(id)'
        ).execute()
        files = results.get('files', [])

        # 2. Update or Create
        if files:
            service.files().update(fileId=files[0]['id'], media_body=media).execute()
            return "updated"
        else:
            meta = {'name': filename, 'parents': ['appDataFolder']}
            service.files().create(body=meta, media_body=media).execute()
            return "created"

    async def save_current_note():
        # 1. Get Data
        vid = get_vid()
        content = notepad.textarea.value
        filename = get_filename(vid)
        
        # 2. Show Loading Spinner
        # (This prevents the user from clicking save twice)
        n = ui.notification('Saving to Cloud...', timeout=None, spinner=True)
        
        try:
            # 3. Run the heavy network call in a separate thread
            # 'service' is passed explicitly to be thread-safe
            result = await run.io_bound(
                save_note_sync, 
                service, 
                vid, 
                content, 
                filename
            )
            
            # 4. Update Index & UI (Back on the main thread)
            index_mgr.add_verse(vid)
            app.storage.user['cached_index'] = index_mgr.data
            
            n.dismiss()
            ui.notify('Saved successfully!', type='positive')
            refresh_ui()
            
        except Exception as e:
            n.dismiss()
            ui.notify(f"Error saving: {e}", type='negative')

    def delete_current_note():
        vid = get_vid()
        try:
            filename = get_filename(vid)
            results = service.files().list(
                q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                spaces='appDataFolder',
                fields='files(id)'
            ).execute()
            files = results.get('files', [])

            if files:
                service.files().delete(fileId=files[0]['id']).execute()
                index_mgr.remove_verse(vid)
                app.storage.user['cached_index'] = index_mgr.data
                notepad.textarea.value = ""
                ui.notify('Note deleted.')
                refresh_ui()
            else:
                ui.notify('Nothing to delete.')
        except Exception as e:
            ui.notify(f"Delete error: {e}", type='negative')

    async def run_rebuild():
        # Run the rebuild
        n = ui.notification('Rebuilding search index...', timeout=None, spinner=True)
        count = await run.io_bound(index_mgr.rebuild_index)
        n.dismiss()
        if count >= 0:
            app.storage.user['cached_index'] = index_mgr.data # Update session
            refresh_ui() # Refresh stats
            ui.notify(f'Search index rebuilt! Found {count} notes.', type='positive')
        else:
            ui.notify('Rebuild failed. Check console logs.', type='negative')

    def download_index_backup():
        """Generates a downloadable JSON file of the current index."""
        if not index_mgr.data:
            ui.notify('Index is empty!', type='warning')
            return
        try:
            # 1. Convert index data to pretty JSON string
            json_str = json.dumps(index_mgr.data, indent=2)
            # 2. Encode it for the browser (Data URI format)
            encoded_json = urllib.parse.quote(json_str)
            # 3. Trigger download
            ui.download(f'data:application/json,{encoded_json}', 'bible_index_backup.json')
            ui.notify('Download started.')
        except Exception as e:
            ui.notify(f'Download error: {e}', type='negative')

    # ---------------------------------------------------------
    # Import a single json or a zip
    # ---------------------------------------------------------

    def process_import_sync(file_bytes, is_zip, drive_service):
        """
        Parses the uploaded bytes (Zip or JSON) and uploads them to Drive.
        Returns a list of successfully imported verse_ids.
        """
        imported_ids = []
        files_to_process = {} # {filename: json_content_dict}

        try:
            # A. Parse the Input (Zip or Single JSON)
            if is_zip:
                with zipfile.ZipFile(io.BytesIO(file_bytes)) as z:
                    for filename in z.namelist():
                        if filename.endswith('.json') and filename != 'bible_index.json':
                            try:
                                content = z.read(filename)
                                data = json.loads(content)
                                files_to_process[filename] = data
                            except:
                                print(f"Skipping invalid file: {filename}")
            else:
                # Single JSON file
                try:
                    data = json.loads(file_bytes)
                    # Quick check if it's a note file
                    if 'verse_id' in data:
                        filename = f"{data['verse_id']}.json"
                        files_to_process[filename] = data
                except:
                    pass

            # B. Upload Loop
            # Note: This loops through every note. For 100+ notes, this takes time.
            for filename, data in files_to_process.items():
                verse_id = data.get('verse_id')
                if not verse_id: continue

                # Prepare content
                # We strictly re-dump the data to ensure clean JSON
                media = MediaIoBaseUpload(
                    io.BytesIO(json.dumps(data).encode('utf-8')), 
                    mimetype='application/json'
                )
                
                # 1. Check if file exists in Drive
                results = drive_service.files().list(
                    q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                    spaces='appDataFolder',
                    fields='files(id)'
                ).execute()
                existing_files = results.get('files', [])

                if existing_files:
                    # Update existing
                    file_id = existing_files[0]['id']
                    drive_service.files().update(fileId=file_id, media_body=media).execute()
                else:
                    # Create new
                    meta = {'name': filename, 'parents': ['appDataFolder']}
                    drive_service.files().create(body=meta, media_body=media).execute()
                
                imported_ids.append(verse_id)

        except Exception as e:
            print(f"Import process failed: {e}")
        
        return imported_ids

    async def handle_import(e):
        """
        UI Callback for the upload button.
        """
        #print(dir(e))
        # 1. Check login
        if not service:
            ui.notify('Please log in to Google first!', type='warning')
            return

        n = ui.notification('Processing Import... Please wait.', timeout=None, spinner=True)
        
        try:
            # 2. Read file bytes
            file_bytes = await e.file.read() 
            filename = getattr(e.file, 'name', 'unknown.json').lower()
            #print(filename)
            is_zip = filename.endswith('.zip')

            # Run background worker
            imported_ids = await run.io_bound(
                process_import_sync, 
                file_bytes, 
                is_zip, 
                service
            )

            # 4. Update Local Index
            if imported_ids:
                count = len(imported_ids)
                for vid in imported_ids:
                    index_mgr.add_verse(vid) # Updates local memory
                
                # Sync index to Drive once at the end
                index_mgr.save_to_drive() 
                app.storage.user['cached_index'] = index_mgr.data
                
                ui.notify(f'Successfully imported {count} notes!', type='positive')
                refresh_ui() # Update your stats/counters
                change_note(book=b, chapter=c, verse=v)
            else:
                ui.notify('No valid notes found to import.', type='warning')

        except Exception as err:
            ui.notify(f'Import failed: {err}', type='negative')
        finally:
            n.dismiss()

    # ---------------------------------------------------------
    # Download a single (the current) note
    # ---------------------------------------------------------
    
    def download_current_note():
        """
        Exports the currently visible note as a JSON file.
        This format is compatible with the 'Import' feature.
        """
        vid = get_vid() # Uses your existing helper
        content = notepad.textarea.value or ''
        
        if not content.strip():
            ui.notify('Cannot download an empty note.', type='warning')
            return

        try:
            # 1. Structure the data exactly like it is stored in Drive
            note_data = {
                "verse_id": vid,
                "content": content,
                "exported_at": "true" # Optional flag
            }
            
            # 2. Convert to JSON bytes
            json_str = json.dumps(note_data, indent=2)
            json_bytes = json_str.encode('utf-8')
            
            # 3. Encode to Base64 (Safe for all browsers/characters)
            b64_str = base64.b64encode(json_bytes).decode()
            
            # 4. Trigger Download
            filename = f"{vid}.json"
            ui.download(f'data:application/json;base64,{b64_str}', filename)
            ui.notify(f'Downloaded {filename}')
            
        except Exception as e:
            ui.notify(f'Download error: {e}', type='negative')

    # ---------------------------------------------------------
    # 1. The Heavy Lifter (Download ALL)
    # ---------------------------------------------------------
    def generate_zip_sync(data, drive_service):
        """
        Synchronous function to download files and create zip.
        Note: We pass 'data' and 'drive_service' explicitly.
        """
        zip_buffer = io.BytesIO()
        
        with zipfile.ZipFile(zip_buffer, 'w', zipfile.ZIP_DEFLATED) as zip_file:
            # Add Index File
            zip_file.writestr('bible_index.json', json.dumps(data, indent=2))
            
            # Loop through notes
            for verse_id in data:
                try:
                    filename = f"{verse_id}.json" # Recreating get_filename logic here for safety
                    
                    # Search logic
                    results = drive_service.files().list(
                        q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                        spaces='appDataFolder',
                        fields='files(id)'
                    ).execute()
                    files = results.get('files', [])

                    if files:
                        file_id = files[0]['id']
                        request = drive_service.files().get_media(fileId=file_id)
                        file_content = request.execute()
                        zip_file.writestr(filename, file_content)
                except Exception as e:
                    print(f"Error zipping {verse_id}: {e}")

        # Finalize
        zip_buffer.seek(0)
        return base64.b64encode(zip_buffer.getvalue()).decode()

    # ---------------------------------------------------------
    # 2. The UI Function (Runs on main thread)
    # ---------------------------------------------------------
    async def download_all_notes_zip():
        if not index_mgr.data:
            ui.notify('No notes to download!', type='warning')
            return

        # 1. Show Spinner (Now it will actually render!)
        n = ui.notification('Downloading notes from Google Drive... Please wait.', timeout=None, spinner=True)
        
        try:
            # 2. Offload the heavy work to a background thread
            # This allows the UI to keep running while Python waits for the result
            b64_zip = await run.io_bound(generate_zip_sync, index_mgr.data, service)
            
            # 3. Trigger Download
            ui.download(f'data:application/zip;base64,{b64_zip}', 'biblemate_notes_backup.zip')
            
            n.dismiss()
            ui.notify('Download started!', type='positive')

        except Exception as e:
            n.dismiss()
            ui.notify(f'Zip error: {e}', type='negative')
    
    # Dialog to confirm deleting a note
    with ui.dialog() as delete_dialog, ui.card():
        ui.label('Are you sure you want to delete this note?')
        with ui.row().classes('justify-end w-full'):
            ui.button(get_translation("Cancel"), on_click=delete_dialog.close).props('flat text-color=secondary')
            ui.button('Delete', color='red', on_click=lambda: (delete_current_note(), delete_dialog.close()))

    # Bible Selection menu
    def additional_items():
        nonlocal gui, bible_selector, area
        with ui.button(icon='more_vert').props(f'flat round color={"white" if app.storage.user["dark_mode"] else "black"}'):
            with ui.menu():
                ui.menu_item(f'👁️ {get_translation("Read")} / {get_translation("Edit")}', on_click=notepad.toggle_mode)
                ui.separator()
                ui.menu_item(f'💾 {get_translation("Save")}', on_click=save_current_note)
                ui.menu_item(f'❌ {get_translation("Delete")}', on_click=delete_dialog.open)
                ui.menu_item(f'📥 {get_translation("Download")}', on_click=download_current_note)
                ui.menu_item(f'📥 {get_translation("Download All")}', on_click=download_all_notes_zip)
                ui.separator()
                with ui.menu_item(f'📤 {get_translation("Import")}', on_click=lambda: upload_element.run_method('pickFiles')):
                    ui.tooltip('Import a single note (*.json file) or a zip file containing multiple notes.')
                ui.separator()
                with ui.menu_item(f'🛠️ {get_translation("Rebuild Index")}', on_click=run_rebuild):
                    ui.tooltip('Rebuild search index for maintenance or repair.')
                ui.menu_item(f'📥 {get_translation("Download Index")}', on_click=download_index_backup)
                ui.separator()
                with ui.menu_item(f'🔒 {get_translation("Logout")}', on_click=lambda: (app.storage.user.clear(), ui.navigate.to('/'))):
                    ui.tooltip('Log out to keep your account secure, especially on shared devices.')
    bible_selector.create_ui(bt, b, c, v, additional_items=additional_items, show_versions=False)

    upload_element = ui.upload(
        on_upload=handle_import, 
        auto_upload=True
    ).props('accept=.json,.zip').classes('hidden')

    notepad.setup_ui()
    def load_initial_content():
        notepad.textarea.value = load_current_note()
    ui.timer(0, lambda: loading(load_initial_content), once=True)




import requests, argparse, os
from rich.markdown import Markdown
from rich.console import Console

parser = argparse.ArgumentParser(description = f"""BibleMate AI Web API Client Options""")
parser.add_argument("default", nargs="*", default=None, help="user query")
parser.add_argument("-l", "--language", action="store", dest="language", help="language option; `eng` by defatul; set this option to `tc` or `sc` to override")
parser.add_argument("-t", "--token", action="store", dest="token", help=f"custom token to get acccess to custom data")
parser.add_argument("-u", "--url", action="store", dest="url", help=f"custom token to get acccess to custom data")
args = parser.parse_args()

def main():
    # 1. Define the URL (ensure your BibleMate Web is running)
    url = args.url if args.url else os.getenv("BM_API_ENDPOINT", "http://localhost:33355/api/data")
    # 2. Define your parameters
    payload = {
        "query": " ".join(args.default) if args.default else ".help", # Required
        "language": args.language if args.language else "eng", # Optional
        "token": args.token if args.token else os.getenv("BM_API_CUSTOM_KEY", "") # Optional
    }
    # 3. Send the GET request
    try:
        response = requests.get(url, params=payload)
        
        # 4. Check if the request was successful (Status Code 200)
        if response.status_code == 200:
            data = response.json()  # Convert JSON response to Python dict
            #print("Success!")
            #print(f"Response: {data}")
            api_content = data.get("content", "[NO_CONTENT]")
            if api_content == "\n\n":
                api_content = "[NO_CONTENT]"
            Console().print(Markdown(api_content))
        else:
            print(f"Error: {response.status_code}")
            #print(response.text)
    except requests.exceptions.ConnectionError:
        print("Could not connect to BibleMate API Server. Is it running?")

if __name__ == "__main__":
    main()


TOOLS_SCHEMA = {'anyalyze_psalms': {'description': 'analyze the context and background of the '
                                    'Psalms in the bible; Psalm reference must '
                                    'be given, e.g. Psalm 23:1-3',
                     'name': 'anyalyze_psalms',
                     'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                         'type': 'object'},
                                                               'type': 'array'}},
                                    'required': ['request'],
                                    'type': 'object'}},
 'ask_bible_scholar': {'description': 'ask a bible scholar about the bible',
                       'name': 'ask_bible_scholar',
                       'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                           'type': 'object'},
                                                                 'type': 'array'}},
                                      'required': ['request'],
                                      'type': 'object'}},
 'ask_pastor': {'description': 'ask a church pastor about the bible',
                'name': 'ask_pastor',
                'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                    'type': 'object'},
                                                          'type': 'array'}},
                               'required': ['request'],
                               'type': 'object'}},
 'ask_theologian': {'description': 'ask a theologian about the bible',
                    'name': 'ask_theologian',
                    'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                        'type': 'object'},
                                                              'type': 'array'}},
                                   'required': ['request'],
                                   'type': 'object'}},
 'compare_bible_translations': {'description': 'compare Bible translations; '
                                               'bible verse reference(s) must '
                                               'be given',
                                'name': 'compare_bible_translations',
                                'parameters': {'properties': {'request': {'type': 'string'}},
                                               'required': ['request'],
                                               'type': 'object'}},
 'explain_bible_meaning': {'description': 'Explain the meaning of the '
                                          'user-given content in reference to '
                                          'the Bible',
                           'name': 'explain_bible_meaning',
                           'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                               'type': 'object'},
                                                                     'type': 'array'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'expound_bible_topic': {'description': 'Expound the user-given topic in '
                                        'reference to the Bible; a topic must '
                                        'be given',
                         'name': 'expound_bible_topic',
                         'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                             'type': 'object'},
                                                                   'type': 'array'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'identify_bible_keywords': {'description': 'Identify bible key words from the '
                                            'user-given content',
                             'name': 'identify_bible_keywords',
                             'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                 'type': 'object'},
                                                                       'type': 'array'}},
                                            'required': ['request'],
                                            'type': 'object'}},
 'interpret_new_testament_verse': {'description': 'Interpret the user-given '
                                                  'bible verse from the New '
                                                  'Testament in the light of '
                                                  'its context, together with '
                                                  'insights of biblical Greek '
                                                  'studies; a new testament '
                                                  'bible verse / reference(s) '
                                                  'must be given',
                                   'name': 'interpret_new_testament_verse',
                                   'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                       'type': 'object'},
                                                                             'type': 'array'}},
                                                  'required': ['request'],
                                                  'type': 'object'}},
 'interpret_old_testament_verse': {'description': 'Interpret the user-given '
                                                  'bible verse from the Old '
                                                  'Testament in the light of '
                                                  'its context, together with '
                                                  'insights of biblical Hebrew '
                                                  'studies; an old testament '
                                                  'bible verse / reference(s) '
                                                  'must be given',
                                   'name': 'interpret_old_testament_verse',
                                   'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                       'type': 'object'},
                                                                             'type': 'array'}},
                                                  'required': ['request'],
                                                  'type': 'object'}},
 'quote_bible_promises': {'description': 'Quote relevant Bible promises in '
                                         'response to user request',
                          'name': 'quote_bible_promises',
                          'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                              'type': 'object'},
                                                                    'type': 'array'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'quote_bible_verses': {'description': 'quote multiple bible verses in '
                                       'response to user request',
                        'name': 'quote_bible_verses',
                        'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                            'type': 'object'},
                                                                  'type': 'array'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'read_bible_commentary': {'description': 'read bible commentary on individual '
                                          'bible verses; bible verse '
                                          'reference(s) must be given, like , '
                                          'like John 3:16 or John 3:16-18',
                           'name': 'read_bible_commentary',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'refine_bible_translation': {'description': 'refine the translation of a '
                                             'Bible verse or passage',
                              'name': 'refine_bible_translation',
                              'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                  'type': 'object'},
                                                                        'type': 'array'}},
                                             'required': ['request'],
                                             'type': 'object'}},
 'retrieve_bible_chapter': {'description': 'retrieve a whole Bible chapter; '
                                           'bible chapter reference must be '
                                           'given, e.g. John 3',
                            'name': 'retrieve_bible_chapter',
                            'parameters': {'properties': {'request': {'type': 'string'}},
                                           'required': ['request'],
                                           'type': 'object'}},
 'retrieve_bible_cross_references': {'description': 'retrieve cross-references '
                                                    'of Bible verses; bible '
                                                    'verse reference(s) must '
                                                    'be given',
                                     'name': 'retrieve_bible_cross_references',
                                     'parameters': {'properties': {'request': {'type': 'string'}},
                                                    'required': ['request'],
                                                    'type': 'object'}},
 'retrieve_bible_verses': {'description': 'retrieve Bible verses; bible verse '
                                          'reference(s) must be given, e.g. '
                                          'John 3:16-17; single or multiple '
                                          'references accepted, e.g. Deut 6:4; '
                                          'Gen 1:26-27',
                           'name': 'retrieve_bible_verses',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'retrieve_hebrew_or_greek_bible_verses': {'description': 'retrieve Hebrew or '
                                                          'Greek Bible verses; '
                                                          'bible verse '
                                                          'reference(s) must '
                                                          'be given, e.g. John '
                                                          '3:16-17; single or '
                                                          'multiple references '
                                                          'accepted, e.g. Deut '
                                                          '6:4; Gen 1:26-27',
                                           'name': 'retrieve_hebrew_or_greek_bible_verses',
                                           'parameters': {'properties': {'request': {'type': 'string'}},
                                                          'required': ['request'],
                                                          'type': 'object'}},
 'retrieve_interlinear_hebrew_or_greek_bible_verses': {'description': 'retrieve '
                                                                      'interlinear '
                                                                      'Hebrew-English '
                                                                      'or '
                                                                      'Greek-English '
                                                                      'Bible '
                                                                      'verses; '
                                                                      'bible '
                                                                      'verse '
                                                                      'reference(s) '
                                                                      'must be '
                                                                      'given, '
                                                                      'e.g. '
                                                                      'John '
                                                                      '3:16-17; '
                                                                      'single '
                                                                      'or '
                                                                      'multiple '
                                                                      'references '
                                                                      'accepted, '
                                                                      'e.g. '
                                                                      'Deut '
                                                                      '6:4; '
                                                                      'Gen '
                                                                      '1:26-27',
                                                       'name': 'retrieve_interlinear_hebrew_or_greek_bible_verses',
                                                       'parameters': {'properties': {'request': {'type': 'string'}},
                                                                      'required': ['request'],
                                                                      'type': 'object'}},
 'retrieve_verse_morphology': {'description': 'retrieve parsing and morphology '
                                              'of individual bible verses; '
                                              'bible verse reference(s) must '
                                              'be given, e.g. John 3:16-17; '
                                              'single or multiple references '
                                              'accepted, e.g. Deut 6:4; Gen '
                                              '1:26-27',
                               'name': 'retrieve_verse_morphology',
                               'parameters': {'properties': {'request': {'type': 'string'}},
                                              'required': ['request'],
                                              'type': 'object'}},
 'search_1_chronicles_only': {'description': 'search the book of 1 Chronicles '
                                             'only; search string must be '
                                             'given',
                              'name': 'search_1_chronicles_only',
                              'parameters': {'properties': {'request': {'type': 'string'}},
                                             'required': ['request'],
                                             'type': 'object'}},
 'search_1_corinthians_only': {'description': 'search the book of 1 '
                                              'Corinthians only; search string '
                                              'must be given',
                               'name': 'search_1_corinthians_only',
                               'parameters': {'properties': {'request': {'type': 'string'}},
                                              'required': ['request'],
                                              'type': 'object'}},
 'search_1_john_only': {'description': 'search the book of 1 John only; search '
                                       'string must be given',
                        'name': 'search_1_john_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_1_kings_only': {'description': 'search the book of 1 Kings only; '
                                        'search string must be given',
                         'name': 'search_1_kings_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_1_peter_only': {'description': 'search the book of 1 Peter only; '
                                        'search string must be given',
                         'name': 'search_1_peter_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_1_samuel_only': {'description': 'search the book of 1 Samuel only; '
                                         'search string must be given',
                          'name': 'search_1_samuel_only',
                          'parameters': {'properties': {'request': {'type': 'string'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'search_1_thessalonians_only': {'description': 'search the book of 1 '
                                                'Thessalonians only; search '
                                                'string must be given',
                                 'name': 'search_1_thessalonians_only',
                                 'parameters': {'properties': {'request': {'type': 'string'}},
                                                'required': ['request'],
                                                'type': 'object'}},
 'search_1_timothy_only': {'description': 'search the book of 1 Timothy only; '
                                          'search string must be given',
                           'name': 'search_1_timothy_only',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'search_2_chronicles_only': {'description': 'search the book of 2 Chronicles '
                                             'only; search string must be '
                                             'given',
                              'name': 'search_2_chronicles_only',
                              'parameters': {'properties': {'request': {'type': 'string'}},
                                             'required': ['request'],
                                             'type': 'object'}},
 'search_2_corinthians_only': {'description': 'search the book of 2 '
                                              'Corinthians only; search string '
                                              'must be given',
                               'name': 'search_2_corinthians_only',
                               'parameters': {'properties': {'request': {'type': 'string'}},
                                              'required': ['request'],
                                              'type': 'object'}},
 'search_2_john_only': {'description': 'search the book of 2 John only; search '
                                       'string must be given',
                        'name': 'search_2_john_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_2_kings_only': {'description': 'search the book of 2 Kings only; '
                                        'search string must be given',
                         'name': 'search_2_kings_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_2_peter_only': {'description': 'search the book of 2 Peter only; '
                                        'search string must be given',
                         'name': 'search_2_peter_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_2_samuel_only': {'description': 'search the book of 2 Samuel only; '
                                         'search string must be given',
                          'name': 'search_2_samuel_only',
                          'parameters': {'properties': {'request': {'type': 'string'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'search_2_thessalonians_only': {'description': 'search the book of 2 '
                                                'Thessalonians only; search '
                                                'string must be given',
                                 'name': 'search_2_thessalonians_only',
                                 'parameters': {'properties': {'request': {'type': 'string'}},
                                                'required': ['request'],
                                                'type': 'object'}},
 'search_2_timothy_only': {'description': 'search the book of 2 Timothy only; '
                                          'search string must be given',
                           'name': 'search_2_timothy_only',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'search_3_john_only': {'description': 'search the book of 3 John only; search '
                                       'string must be given',
                        'name': 'search_3_john_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_acts_only': {'description': 'search the book of Acts only; search '
                                     'string must be given',
                      'name': 'search_acts_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_amos_only': {'description': 'search the book of Amos only; search '
                                     'string must be given',
                      'name': 'search_amos_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_colossians_only': {'description': 'search the book of Colossians '
                                           'only; search string must be given',
                            'name': 'search_colossians_only',
                            'parameters': {'properties': {'request': {'type': 'string'}},
                                           'required': ['request'],
                                           'type': 'object'}},
 'search_daniel_only': {'description': 'search the book of Daniel only; search '
                                       'string must be given',
                        'name': 'search_daniel_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_deuteronomy_only': {'description': 'search the book of Deuteronomy '
                                            'only; search string must be given',
                             'name': 'search_deuteronomy_only',
                             'parameters': {'properties': {'request': {'type': 'string'}},
                                            'required': ['request'],
                                            'type': 'object'}},
 'search_ecclesiastes_only': {'description': 'search the book of Ecclesiastes '
                                             'only; search string must be '
                                             'given',
                              'name': 'search_ecclesiastes_only',
                              'parameters': {'properties': {'request': {'type': 'string'}},
                                             'required': ['request'],
                                             'type': 'object'}},
 'search_ephesians_only': {'description': 'search the book of Ephesians only; '
                                          'search string must be given',
                           'name': 'search_ephesians_only',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'search_esther_only': {'description': 'search the book of Esther only; search '
                                       'string must be given',
                        'name': 'search_esther_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_exodus_only': {'description': 'search the book of Exodus only; search '
                                       'string must be given',
                        'name': 'search_exodus_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_ezekiel_only': {'description': 'search the book of Ezekiel only; '
                                        'search string must be given',
                         'name': 'search_ezekiel_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_ezra_only': {'description': 'search the book of Ezra only; search '
                                     'string must be given',
                      'name': 'search_ezra_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_galatians_only': {'description': 'search the book of Galatians only; '
                                          'search string must be given',
                           'name': 'search_galatians_only',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'search_genesis_only': {'description': 'search the book of Genesis only; '
                                        'search string must be given',
                         'name': 'search_genesis_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_habakkuk_only': {'description': 'search the book of Habakkuk only; '
                                         'search string must be given',
                          'name': 'search_habakkuk_only',
                          'parameters': {'properties': {'request': {'type': 'string'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'search_haggai_only': {'description': 'search the book of Haggai only; search '
                                       'string must be given',
                        'name': 'search_haggai_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_hebrews_only': {'description': 'search the book of Hebrews only; '
                                        'search string must be given',
                         'name': 'search_hebrews_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_hosea_only': {'description': 'search the book of Hosea only; search '
                                      'string must be given',
                       'name': 'search_hosea_only',
                       'parameters': {'properties': {'request': {'type': 'string'}},
                                      'required': ['request'],
                                      'type': 'object'}},
 'search_isaiah_only': {'description': 'search the book of Isaiah only; search '
                                       'string must be given',
                        'name': 'search_isaiah_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_james_only': {'description': 'search the book of James only; search '
                                      'string must be given',
                       'name': 'search_james_only',
                       'parameters': {'properties': {'request': {'type': 'string'}},
                                      'required': ['request'],
                                      'type': 'object'}},
 'search_jeremiah_only': {'description': 'search the book of Jeremiah only; '
                                         'search string must be given',
                          'name': 'search_jeremiah_only',
                          'parameters': {'properties': {'request': {'type': 'string'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'search_job_only': {'description': 'search the book of Job only; search '
                                    'string must be given',
                     'name': 'search_job_only',
                     'parameters': {'properties': {'request': {'type': 'string'}},
                                    'required': ['request'],
                                    'type': 'object'}},
 'search_joel_only': {'description': 'search the book of Joel only; search '
                                     'string must be given',
                      'name': 'search_joel_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_john_only': {'description': 'search the book of John only; search '
                                     'string must be given',
                      'name': 'search_john_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_jonah_only': {'description': 'search the book of Jonah only; search '
                                      'string must be given',
                       'name': 'search_jonah_only',
                       'parameters': {'properties': {'request': {'type': 'string'}},
                                      'required': ['request'],
                                      'type': 'object'}},
 'search_joshua_only': {'description': 'search the book of Joshua only; search '
                                       'string must be given',
                        'name': 'search_joshua_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_jude_only': {'description': 'search the book of Jude only; search '
                                     'string must be given',
                      'name': 'search_jude_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_judges_only': {'description': 'search the book of Judges only; search '
                                       'string must be given',
                        'name': 'search_judges_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_lamentations_only': {'description': 'search the book of Lamentations '
                                             'only; search string must be '
                                             'given',
                              'name': 'search_lamentations_only',
                              'parameters': {'properties': {'request': {'type': 'string'}},
                                             'required': ['request'],
                                             'type': 'object'}},
 'search_leviticus_only': {'description': 'search the book of Leviticus only; '
                                          'search string must be given',
                           'name': 'search_leviticus_only',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'search_luke_only': {'description': 'search the book of Luke only; search '
                                     'string must be given',
                      'name': 'search_luke_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_malachi_only': {'description': 'search the book of Malachi only; '
                                        'search string must be given',
                         'name': 'search_malachi_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_mark_only': {'description': 'search the book of Mark only; search '
                                     'string must be given',
                      'name': 'search_mark_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_matthew_only': {'description': 'search the book of Matthew only; '
                                        'search string must be given',
                         'name': 'search_matthew_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_micah_only': {'description': 'search the book of Micah only; search '
                                      'string must be given',
                       'name': 'search_micah_only',
                       'parameters': {'properties': {'request': {'type': 'string'}},
                                      'required': ['request'],
                                      'type': 'object'}},
 'search_nahum_only': {'description': 'search the book of Nahum only; search '
                                      'string must be given',
                       'name': 'search_nahum_only',
                       'parameters': {'properties': {'request': {'type': 'string'}},
                                      'required': ['request'],
                                      'type': 'object'}},
 'search_nehemiah_only': {'description': 'search the book of Nehemiah only; '
                                         'search string must be given',
                          'name': 'search_nehemiah_only',
                          'parameters': {'properties': {'request': {'type': 'string'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'search_numbers_only': {'description': 'search the book of Numbers only; '
                                        'search string must be given',
                         'name': 'search_numbers_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_obadiah_only': {'description': 'search the book of Obadiah only; '
                                        'search string must be given',
                         'name': 'search_obadiah_only',
                         'parameters': {'properties': {'request': {'type': 'string'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'search_philemon_only': {'description': 'search the book of Philemon only; '
                                         'search string must be given',
                          'name': 'search_philemon_only',
                          'parameters': {'properties': {'request': {'type': 'string'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'search_philippians_only': {'description': 'search the book of Philippians '
                                            'only; search string must be given',
                             'name': 'search_philippians_only',
                             'parameters': {'properties': {'request': {'type': 'string'}},
                                            'required': ['request'],
                                            'type': 'object'}},
 'search_proverbs_only': {'description': 'search the book of Proverbs only; '
                                         'search string must be given',
                          'name': 'search_proverbs_only',
                          'parameters': {'properties': {'request': {'type': 'string'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'search_psalms_only': {'description': 'search the book of Psalms only; search '
                                       'string must be given',
                        'name': 'search_psalms_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_revelation_only': {'description': 'search the book of Revelation '
                                           'only; search string must be given',
                            'name': 'search_revelation_only',
                            'parameters': {'properties': {'request': {'type': 'string'}},
                                           'required': ['request'],
                                           'type': 'object'}},
 'search_romans_only': {'description': 'search the book of Romans only; search '
                                       'string must be given',
                        'name': 'search_romans_only',
                        'parameters': {'properties': {'request': {'type': 'string'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'search_ruth_only': {'description': 'search the book of Ruth only; search '
                                     'string must be given',
                      'name': 'search_ruth_only',
                      'parameters': {'properties': {'request': {'type': 'string'}},
                                     'required': ['request'],
                                     'type': 'object'}},
 'search_song_of_songs_only': {'description': 'search the book of Song of '
                                              'Songs only; search string must '
                                              'be given',
                               'name': 'search_song_of_songs_only',
                               'parameters': {'properties': {'request': {'type': 'string'}},
                                              'required': ['request'],
                                              'type': 'object'}},
 'search_the_whole_bible': {'description': 'search the whole bible; search '
                                           'string must be given',
                            'name': 'search_the_whole_bible',
                            'parameters': {'properties': {'request': {'type': 'string'}},
                                           'required': ['request'],
                                           'type': 'object'}},
 'search_titus_only': {'description': 'search the book of Titus only; search '
                                      'string must be given',
                       'name': 'search_titus_only',
                       'parameters': {'properties': {'request': {'type': 'string'}},
                                      'required': ['request'],
                                      'type': 'object'}},
 'search_zechariah_only': {'description': 'search the book of Zechariah only; '
                                          'search string must be given',
                           'name': 'search_zechariah_only',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'search_zephaniah_only': {'description': 'search the book of Zephaniah only; '
                                          'search string must be given',
                           'name': 'search_zephaniah_only',
                           'parameters': {'properties': {'request': {'type': 'string'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'study_bible_themes': {'description': 'Study Bible Themes in relation to the '
                                       'user content',
                        'name': 'study_bible_themes',
                        'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                            'type': 'object'},
                                                                  'type': 'array'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'study_new_testament_themes': {'description': 'Study Bible Themes in a New '
                                               'Testament passage; new '
                                               'testament bible book / chapter '
                                               '/ passage / reference(s) must '
                                               'be given',
                                'name': 'study_new_testament_themes',
                                'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                    'type': 'object'},
                                                                          'type': 'array'}},
                                               'required': ['request'],
                                               'type': 'object'}},
 'study_old_testament_themes': {'description': 'Study Bible Themes in a Old '
                                               'Testament passage; old '
                                               'testatment bible book / '
                                               'chapter / passage / '
                                               'reference(s) must be given',
                                'name': 'study_old_testament_themes',
                                'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                    'type': 'object'},
                                                                          'type': 'array'}},
                                               'required': ['request'],
                                               'type': 'object'}},
 'translate_greek_bible_verse': {'description': 'Translate a Greek bible '
                                                'verse: Greek bible text must '
                                                'be given',
                                 'name': 'translate_greek_bible_verse',
                                 'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                     'type': 'object'},
                                                                           'type': 'array'}},
                                                'required': ['request'],
                                                'type': 'object'}},
 'translate_hebrew_bible_verse': {'description': 'Translate a Hebrew bible '
                                                 'verse; Hebrew bible text '
                                                 'must be given',
                                  'name': 'translate_hebrew_bible_verse',
                                  'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                      'type': 'object'},
                                                                            'type': 'array'}},
                                                 'required': ['request'],
                                                 'type': 'object'}},
 'write_bible_applications': {'description': 'Provide detailed applications of '
                                             'a bible passages; bible book / '
                                             'chapter / passage / reference(s) '
                                             'must be given',
                              'name': 'write_bible_applications',
                              'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                  'type': 'object'},
                                                                        'type': 'array'}},
                                             'required': ['request'],
                                             'type': 'object'}},
 'write_bible_book_introduction': {'description': 'Write a detailed '
                                                  'introduction on a book in '
                                                  'the bible; bible book must '
                                                  'be given',
                                   'name': 'write_bible_book_introduction',
                                   'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                       'type': 'object'},
                                                                             'type': 'array'}},
                                                  'required': ['request'],
                                                  'type': 'object'}},
 'write_bible_canonical_context': {'description': 'Write about canonical '
                                                  'context of a bible book / '
                                                  'chapter / passage; bible '
                                                  'book / chapter / passage / '
                                                  'reference(s) must be given',
                                   'name': 'write_bible_canonical_context',
                                   'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                       'type': 'object'},
                                                                             'type': 'array'}},
                                                  'required': ['request'],
                                                  'type': 'object'}},
 'write_bible_chapter_summary': {'description': 'Write a detailed '
                                                'interpretation on a bible '
                                                'chapter; a bible chapter must '
                                                'be given',
                                 'name': 'write_bible_chapter_summary',
                                 'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                     'type': 'object'},
                                                                           'type': 'array'}},
                                                'required': ['request'],
                                                'type': 'object'}},
 'write_bible_character_study': {'description': 'Write comprehensive '
                                                'information on a given bible '
                                                'character in the bible; a '
                                                'bible character name must be '
                                                'given',
                                 'name': 'write_bible_character_study',
                                 'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                     'type': 'object'},
                                                                           'type': 'array'}},
                                                'required': ['request'],
                                                'type': 'object'}},
 'write_bible_devotion': {'description': 'Write a devotion on a bible passage; '
                                         'bible book / chapter / passage / '
                                         'reference(s) must be given',
                          'name': 'write_bible_devotion',
                          'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                              'type': 'object'},
                                                                    'type': 'array'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'write_bible_insights': {'description': 'Write exegetical insights in detail '
                                         'on a bible passage; bible book / '
                                         'chapter / passage / reference(s) '
                                         'must be given',
                          'name': 'write_bible_insights',
                          'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                              'type': 'object'},
                                                                    'type': 'array'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'write_bible_location_study': {'description': 'write comprehensive '
                                               'information on a bible '
                                               'location; a bible location '
                                               'name must be given',
                                'name': 'write_bible_location_study',
                                'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                    'type': 'object'},
                                                                          'type': 'array'}},
                                               'required': ['request'],
                                               'type': 'object'}},
 'write_bible_outline': {'description': 'provide a detailed outline of a bible '
                                        'book / chapter / passage; bible book '
                                        '/ chapter / passage / reference(s) '
                                        'must be given',
                         'name': 'write_bible_outline',
                         'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                             'type': 'object'},
                                                                   'type': 'array'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'write_bible_perspectives': {'description': 'Write biblical perspectives and '
                                             'principles in relation to the '
                                             'user content',
                              'name': 'write_bible_perspectives',
                              'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                  'type': 'object'},
                                                                        'type': 'array'}},
                                             'required': ['request'],
                                             'type': 'object'}},
 'write_bible_prayer': {'description': 'Write a prayer pertaining to the user '
                                       'content in reference to the Bible',
                        'name': 'write_bible_prayer',
                        'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                            'type': 'object'},
                                                                  'type': 'array'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'write_bible_questions': {'description': 'Write thought-provoking questions '
                                          'for bible study group discussion; '
                                          'bible book / chapter / passage / '
                                          'reference(s) must be given',
                           'name': 'write_bible_questions',
                           'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                               'type': 'object'},
                                                                     'type': 'array'}},
                                          'required': ['request'],
                                          'type': 'object'}},
 'write_bible_related_summary': {'description': 'Write a summary on the '
                                                'user-given content in '
                                                'reference to the Bible',
                                 'name': 'write_bible_related_summary',
                                 'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                     'type': 'object'},
                                                                           'type': 'array'}},
                                                'required': ['request'],
                                                'type': 'object'}},
 'write_bible_sermon': {'description': 'Write a bible sermon based on a bible '
                                       'passage; bible book / chapter / '
                                       'passage / reference(s) must be given',
                        'name': 'write_bible_sermon',
                        'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                            'type': 'object'},
                                                                  'type': 'array'}},
                                       'required': ['request'],
                                       'type': 'object'}},
 'write_bible_theology': {'description': 'write the theological messages '
                                         'conveyed in the user-given content, '
                                         'in reference to the Bible',
                          'name': 'write_bible_theology',
                          'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                              'type': 'object'},
                                                                    'type': 'array'}},
                                         'required': ['request'],
                                         'type': 'object'}},
 'write_bible_thought_progression': {'description': 'write Bible Thought '
                                                    'Progression of a bible '
                                                    'book / chapter / passage; '
                                                    'bible book / chapter / '
                                                    'passage / reference(s) '
                                                    'must be given',
                                     'name': 'write_bible_thought_progression',
                                     'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                         'type': 'object'},
                                                                               'type': 'array'}},
                                                    'required': ['request'],
                                                    'type': 'object'}},
 'write_new_testament_highlights': {'description': 'Write Highlights in a New '
                                                   'Testament passage in the '
                                                   'bible; new testament bible '
                                                   'book / chapter / passage / '
                                                   'reference(s) must be given',
                                    'name': 'write_new_testament_highlights',
                                    'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                        'type': 'object'},
                                                                              'type': 'array'}},
                                                   'required': ['request'],
                                                   'type': 'object'}},
 'write_new_testament_historical_context': {'description': 'write the Bible '
                                                           'Historical Context '
                                                           'of a New Testament '
                                                           'passage in the '
                                                           'bible; new '
                                                           'testament bible '
                                                           'book / chapter / '
                                                           'passage / '
                                                           'reference(s) must '
                                                           'be given',
                                            'name': 'write_new_testament_historical_context',
                                            'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                                'type': 'object'},
                                                                                      'type': 'array'}},
                                                           'required': ['request'],
                                                           'type': 'object'}},
 'write_old_testament_highlights': {'description': 'Write Highlights in a Old '
                                                   'Testament passage in the '
                                                   'bible; old testament bible '
                                                   'book / chapter / passage / '
                                                   'reference(s) must be given',
                                    'name': 'write_old_testament_highlights',
                                    'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                        'type': 'object'},
                                                                              'type': 'array'}},
                                                   'required': ['request'],
                                                   'type': 'object'}},
 'write_old_testament_historical_context': {'description': 'write the Bible '
                                                           'Historical Context '
                                                           'of a Old Testament '
                                                           'passage in the '
                                                           'bible; old '
                                                           'testament bible '
                                                           'book / chapter / '
                                                           'passage / '
                                                           'reference(s) must '
                                                           'be given',
                                            'name': 'write_old_testament_historical_context',
                                            'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                                'type': 'object'},
                                                                                      'type': 'array'}},
                                                           'required': ['request'],
                                                           'type': 'object'}},
 'write_pastor_prayer': {'description': 'write a prayer, out of a church '
                                        'pastor heart, based on user input',
                         'name': 'write_pastor_prayer',
                         'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                             'type': 'object'},
                                                                   'type': 'array'}},
                                        'required': ['request'],
                                        'type': 'object'}},
 'write_short_bible_prayer': {'description': 'Write a short prayer, in one '
                                             'paragraph only, pertaining to '
                                             'the user content in reference to '
                                             'the Bible',
                              'name': 'write_short_bible_prayer',
                              'parameters': {'properties': {'request': {'items': {'additionalProperties': True,
                                                                                  'type': 'object'},
                                                                        'type': 'array'}},
                                             'required': ['request'],
                                             'type': 'object'}}}


from nicegui import ui, app
from biblemateweb.translations.eng import translation_eng
from biblemateweb.translations.tc import translation_tc
from biblemateweb.translations.sc import translation_sc


def get_translation(text: str):
    if app.storage.user["ui_language"] == "tc":
        return translation_tc.get(text, text)
    elif app.storage.user["ui_language"] == "sc":
        return translation_sc.get(text, text)
    return translation_eng.get(text, text)

class FilenameDialog(ui.dialog):
    def __init__(self):
        super().__init__()

        card_classes = "w-full max-w-md"

        with self, ui.card().classes(card_classes):

            ui.label(get_translation("Download")).classes('text-xl font-bold')
            ui.label(get_translation("Please enter a name for your file:"))

            # Input field for the filename
            self.filename_input = ui.input(label='Filename', placeholder='example.txt') \
                .classes(card_classes).props('autofocus') \
                .on('keydown.enter', lambda: self.submit(self.filename_input.value))  # Allow pressing Enter to submit

            with ui.row().classes('w-full justify-center shrink-0 pt-4'):
                # Button 1: Cancel
                ui.button(get_translation("Cancel"), on_click=lambda: self.submit(None)) \
                    .props('flat color=grey')
                
                # Button 2: Confirm
                ui.button(get_translation("Confirm"), on_click=lambda: self.submit(self.filename_input.value))

    def open_with_filename(self, filename: str = "BibleMate_AI"):
        """
        Updates the dialog with the provided filename and opens it.
        """
        self.filename_input.value = filename
 
        self.open() # Explicitly open the dialog
        return self


from biblemateweb import config, getLexiconList, get_translation
from nicegui import ui, app, run
import re, apsw
from biblemateweb.data.cr_books import cr_books
from biblemateweb.fx.shared import get_image_data_uri

def fetch_bible_lexicons_entry(client_lexicons, lexicon, topic):
    db = get_lexicon_path(client_lexicons, lexicon)
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        sql_query = f"SELECT Definition FROM Lexicon WHERE Topic=? limit 1"
        cursor.execute(sql_query, (topic,))
        fetch = cursor.fetchone()
        content = fetch[0] if fetch else ""
    return content

def get_lexicon_path(client_lexicons, lexicon_name):
    if not lexicon_name in client_lexicons:
        return client_lexicons[app.storage.user.get('favorite_lexicon', 'Morphology')]
    if lexicon_name in config.lexicons_custom:
        return config.lexicons_custom[lexicon_name]
    elif lexicon_name in config.lexicons:
        return config.lexicons[lexicon_name]

def fetch_all_lexicons_entries(client_lexicons, lexicon):
    db = get_lexicon_path(client_lexicons, lexicon)
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        sql_query = f"SELECT Topic FROM Lexicon"
        cursor.execute(sql_query)
        all_entries = [i[0] for i in cursor.fetchall()]
    return list(set([i for i in all_entries if i]))

def search_bible_lexicons(gui=None, q='', **_):

    last_entry = q
    client_lexicons = getLexiconList(app.storage.client["custom"])

    if q:
        if q.startswith("E") and not app.storage.user['favorite_lexicon'] in ("Morphology", "ConcordanceMorphology", "ConcordanceBook"):
            app.storage.user['favorite_lexicon'] = "Morphology"
        elif q.startswith("BDB"):
            app.storage.user['favorite_lexicon'] = "BDB"
        elif q.startswith("H"):
            app.storage.user['favorite_lexicon'] = app.storage.user.get('hebrew_lexicon', 'Morphology')
        elif q.startswith("G"):
            app.storage.user['favorite_lexicon'] = app.storage.user.get('greek_lexicon', 'Morphology')

    scope_select = None

    def cr(event):
        nonlocal gui
        b, c, v, *_ = event.args
        b = cr_books.get(b, b)
        gui.change_area_1_bible_chapter(None, b, c, v)

    def bcv(event):
        nonlocal gui
        b, c, v, *_ = event.args
        gui.change_area_1_bible_chapter(None, b, c, v)
    
    def website(event):
        url, *_ = event.args
        ui.navigate.to(url, new_tab=True)

    async def bdbid(event):
        nonlocal input_field
        id, *_ = event.args
        input_field.value = id
        await handle_enter(None)

    async def lex(event):
        nonlocal input_field
        id, *_ = event.args
        input_field.value = id
        await handle_enter(None)

    ui.on('bcv', bcv)
    ui.on('cr', cr)
    ui.on('website', website)
    ui.on('bdbid', bdbid)
    ui.on('lex', lex)

    lexicon_module = app.storage.user.get('favorite_lexicon', 'Morphology')
    if lexicon_module not in client_lexicons:
        lexicon_module = 'Morphology'
        app.storage.user['favorite_lexicon'] = lexicon_module

    # ----------------------------------------------------------
    # Core: Fetch and Display
    # ----------------------------------------------------------

    async def change_module(new_module):
        nonlocal input_field, lexicon_module
        lexicon_module = new_module
        app.storage.user['favorite_lexicon'] = new_module
        n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
        all_entries = await run.io_bound(fetch_all_lexicons_entries, client_lexicons, new_module)
        n.dismiss()
        input_field.set_autocomplete(all_entries)
        input_field.props(f'''placeholder="{get_translation('Search')} {new_module} ..."''')
        if scope_select and scope_select.value != new_module:
            scope_select.value = new_module

    def handle_up_arrow():
        nonlocal last_entry, input_field
        if not input_field.value.strip():
            input_field.value = last_entry

    async def handle_enter(_, keep=True):
        nonlocal content_container, gui, input_field, lexicon_module, last_entry

        topic = input_field.value.strip()
        if not topic:
            return

        last_entry = topic
        input_field.disable()

        try:

            # update tab records
            if keep:
                gui.update_active_area2_tab_records(q=topic)

            if (topic.startswith("E") and not lexicon_module in ("Morphology", "ConcordanceMorphology", "ConcordanceBook")):
                await change_module("Morphology")
            elif (topic.startswith("G") and lexicon_module == "BDB"):
                await change_module(app.storage.user.get('greek_lexicon', 'Morphology'))
            elif topic.startswith("BDB") or (topic.startswith("H") and lexicon_module in ("Morphology", "ConcordanceMorphology", "ConcordanceBook")):
                await change_module("BDB")

            n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
            content = await run.io_bound(fetch_bible_lexicons_entry, client_lexicons, lexicon_module, topic)
            n.dismiss()

        except Exception as e:
            # Handle errors (e.g., network failure)
            ui.notify(f'Error: {e}', type='negative')

        finally:
            # ALWAYS re-enable the input, even if an error occurred above
            input_field.enable()
            # Optional: Refocus the cursor so the user can type the next query immediately
            input_field.run_method('focus')

        # Clear existing rows first
        content_container.clear()

        with content_container:
            # Morhology Lexicon
            if lexicon_module == "Morphology":
                content = re.sub(r'''\[<ref onclick="(searchBook|searchCode)\(.*?\)">search</ref>\]''', "", content)
                content = re.sub("(</[on]tgloss>)", r"\1<br>", content)

            # convert links, e.g. <ref onclick="bcv(3,19,26)">
            content = re.sub(r'''(onclick|ondblclick)="(bdbid|lex|cr|bcv|website)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
            content = re.sub(r"""(onclick|ondblclick)='(bdbid|lex|cr|bcv|website)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", content)
            # remove map
            content = content.replace('<div id="map" style="width:100%;height:500px"></div>', "")
            content = re.sub(r'<script.*?>.*?</script>', '', content, flags=re.DOTALL)
            # convert colors for dark mode, e.g. <font color="brown">
            if app.storage.user['dark_mode']:
                content = content.replace('color="brown">', 'color="pink">')
                content = content.replace('color="navy">', 'color="lightskyblue">')
                content = content.replace('<table bgcolor="#BFBFBF"', '<table bgcolor="#424242"')
                content = content.replace('<td bgcolor="#FFFFFF">', '<td bgcolor="#212121">')
                content = content.replace('<tr bgcolor="#FFFFFF">', '<tr bgcolor="#212121">')
                content = content.replace('<tr bgcolor="#DFDFDF">', '<tr bgcolor="#303030">')
            # convert images to data URI
            def replace_img(match):
                img_module = match.group(1)
                img_src = match.group(2)
                img_src = f"{img_module}_{img_src}"
                data_uri = get_image_data_uri(img_module, img_src)
                if data_uri:
                    return f'<img style="display: inline-block;" src="{data_uri}"/>'
                else:
                    return match.group(0)  # return original if not found
            content = re.sub(r'<img src="getImage.php\?resource=([A-Z]+?)&id=(.+?)"/>', replace_img, content)

            # display
            ui.html(f'<div class="content-text">{content}</div>', sanitize=False)

        # Clear input so user can start typing to filter immediately
        if not content:
            ui.notify("No entry found.", color='warning')

    # ==============================================================================
    # 3. UI LAYOUT
    # ==============================================================================
    if q and ":::" in q:
        additional_options, q = q.split(":::", 1)
        if additional_options.strip() in client_lexicons:
            app.storage.user['favorite_lexicon'] = additional_options.strip()

    with ui.row().classes('w-full max-w-3xl mx-auto m-0 py-0 px-4 items-center'):
        scope_select = ui.select(
            options=client_lexicons,
            value=app.storage.user.get('favorite_lexicon', 'Morphology'),
            with_input=True
        ).classes('w-22').props('dense')

        input_field = ui.input(
            autocomplete=[],
            placeholder=f'Search {lexicon_module} ...'
        ).classes('flex-grow text-lg') \
        .props('outlined dense clearable autofocus enterkeyhint="search"')

        input_field.on('keydown.enter.prevent', handle_enter)
        input_field.on('keydown.up', handle_up_arrow)
        with input_field.add_slot('append'):
            ui.icon('history') \
                .on('click', handle_up_arrow) \
                .classes('text-sm cursor-pointer text-secondary').tooltip('Restore last entry')

        async def get_all_entries(lexicon):
            all_entries = await run.io_bound(fetch_all_lexicons_entries, client_lexicons, lexicon)
            input_field.set_autocomplete(all_entries)
        n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
        ui.timer(0, get_all_entries, once=True)
        n.dismiss()

        async def handle_scope_change(e):
            nonlocal lexicon_module
            new_module = e.value
            await change_module(new_module)
            await handle_enter(None)
        scope_select.on_value_change(handle_scope_change)

    # --- Main Content Area ---
    with ui.column().classes('w-full items-center'):
        # Define the container HERE within the layout structure
        content_container = ui.column().classes('w-full transition-all !gap-1')

    if q:
        input_field.value = q
        ui.timer(0, lambda: handle_enter(None, keep=False), once=True)
    else:
        input_field.run_method('focus')


from nicegui import ui
from biblemateweb import get_translation
import traceback, tempfile, pypandoc


class ReviewDialog(ui.dialog):
    def __init__(self):
        super().__init__()
        # 1. flex column: Essential for the card
        with self, ui.card().classes('w-11/12 max-w-none h-[90vh] flex column'):
            
            self.text_label = ui.label(get_translation("Partner Mode Review")).classes('text-h6 font-bold')
            
            # 2. Add the custom CSS class 'biblemate-editor'
            self.editor = ui.textarea(label='Edit Content') \
                .classes('w-full flex-grow biblemate-editor') \
                .props('outlined') 
            
            # The buttons stay at the bottom
            with ui.row().classes('w-full justify-center'):
                ui.button(get_translation("Cancel"), on_click=lambda: self.submit(None)) \
                    .props('flat color=grey')
                ui.button("📥 TXT", on_click=self.download_txt) \
                    .props('flat color=grey')
                ui.button("📥 DOCX", on_click=self.download_docx) \
                    .props('flat color=grey')
                
                ui.button(get_translation("Confirm"), on_click=lambda: self.submit(self.editor.value))

    def open_with_text(self, text_to_review: str, label: str = None):
        self.editor.value = text_to_review
        if label is not None:
            self.text_label.text = label
        return self

    def download_txt(self):
        if content := self.editor.value:
            ui.download(content.encode('utf-8'), 'BibleMate_AI_content.txt')
            ui.notify(get_translation("Downloaded!"), type='positive')
        else:
            ui.notify(get_translation("Nothing to download"), type='negative')

    def download_docx(self):
        if content := self.editor.value:
            try:
                # 1. Create a temporary file that acts as the bridge
                # 'delete=False' is sometimes needed on Windows to close/re-open, 
                # but in a simple flow, we can just read it back.
                with tempfile.NamedTemporaryFile(suffix=".docx") as tmp:
                    
                    # 2. Convert Markdown -> DOCX file
                    pypandoc.convert_text(
                        content, 
                        'docx', 
                        format='md', 
                        outputfile=tmp.name
                    )
                    
                    # 3. Read bytes back into memory
                    tmp.seek(0)
                    docx_bytes = tmp.read()

                # 4. Trigger download in NiceGUI (no file left on server)
                ui.download(docx_bytes, filename='BibleMate_AI_content.docx')
                ui.notify(get_translation("Downloaded!"), type='positive')
            except:
                traceback.print_exc()
        else:
            ui.notify(get_translation("Nothing to download"), type='negative')